from typing import Union
from collections.abc import Iterable

def CNN(input_shape,
        output_shape,
        filters: tuple=(6, 16, 120),
        kernel_sizes: tuple=(3, 3, 5),
        paddings: tuple=(0, 0, 0),
        batch_norm: bool=False,
        dropouts: Union[float, list[float]] = 0.0,
        pool_sizes: tuple=(2, 2, 0),
        fc_units: int=(84, ),
        prepool: int=0,

        dims: int=1):
    import tensorflow as tf
    from tensorflow.keras.models import Sequential
    from tensorflow.keras.layers import Flatten, Dense, BatchNormalization, Dropout, Activation, Input
    from tensorflow.keras.optimizers import Adadelta, SGD, Adam
    from tensorflow.keras.callbacks import LambdaCallback
    import tensorflow.keras.backend as backend

    if dims == 1:
        import qualia_core.learningmodel.keras.layers1d as layers_t
    elif dims == 2:
        import qualia_core.learningmodel.keras.layers2d as layers_t
    else:
        raise ValueError('Only dims=1 or dims=2 supported')

    # Backward compatibility for config not defining dropout as a list
    if not isinstance(dropouts, list):
        dropouts = [dropouts] * (len(filters) + len(fc_units))

    optimizer = Adam()

    model = Sequential()

    # Conv → MaxPool → Conv → MaxPool → Conv → Flatten → FullyConnected → FullyConnected(out)
    # No Dropout/BatchNormalization: doesn't improve test acc on UCI-HAR (even worse)

    model.add(Input(shape=input_shape))

    if prepool:
        model.add(layers_t.AveragePooling(pool_size=prepool))

    for f, ks, ps, padding, dropout in zip(filters, kernel_sizes, pool_sizes, paddings, dropouts):
        if isinstance(padding, Iterable) and sum(padding) > 0 or padding > 0:
            model.add(layers_t.ZeroPadding(padding=padding))
        model.add(layers_t.Conv(filters=f, kernel_size=ks))
        model.add(Activation('relu'))
        if batch_norm:
            model.add(BatchNormalization())
        if dropout:
            model.add(Dropout(dropout))
        if ps: # Optional MaxPool, must specify None in pool_sizes param when no MaxPool layer should be generated after Conv/Act
            model.add(layers_t.MaxPooling(pool_size=ps))

    model.add(Flatten())

    for fc_unit in fc_units:
        model.add(Dense(fc_unit))
        model.add(Activation('relu'))

    # Output classifier
    model.add(Dense(output_shape[0]))
    model.add(Activation('softmax'))

    return model



from __future__ import annotations

import dataclasses
import importlib.util
import logging
import sys
from dataclasses import dataclass
from typing import Final

from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

logger = logging.getLogger(__name__)

import_package_names: Final[list[str]] = ['dataset', 'deployment', 'preprocessing', 'learningframework', 'postprocessing']

@dataclass
class QualiaComponent:
    dataset: ModuleType | None = None
    deployment: ModuleType | None = None
    learningframework: ModuleType | None = None
    postprocessing: ModuleType | None = None
    preprocessing: ModuleType | None = None
    converter: ModuleType | None = None

    def package_names(self) -> tuple[str, ...]:
        return tuple(field.name for field in dataclasses.fields(self) if getattr(self, field.name) is not None)

def import_package_from_plugin(plugin_name: str, package_name: str) -> ModuleType | None:
    if importlib.util.find_spec(f'{plugin_name}.{package_name}') is None:
        logger.info('%s module not found in "%s" plugin', package_name, plugin_name)
        return None

    return importlib.import_module(f'{plugin_name}.{package_name}')

def load_plugin(plugin_name: str) -> QualiaComponent:

    packages = {package_name: import_package_from_plugin(plugin_name=plugin_name, package_name=package_name)
                    for package_name in import_package_names}

    component = QualiaComponent(**packages)
    component.converter = component.postprocessing

    logger.info("Loaded component '%s' with packages %s", plugin_name, component.package_names())

    return component

def load_plugins(plugin_names: list[str]) -> QualiaComponent:
    packages = load_plugin('qualia_core')
    for package_name in import_package_names:
        # Delete qualia_core.* imports from modules cache to force clean reload to prevent breaking direct qualia_core.* imports
        # since we modify their __dict__ below
        del sys.modules[f'qualia_core.{package_name}']
    for plugin_name in plugin_names:
        plugin = load_plugin(plugin_name)
        for package_name in plugin.package_names():
            getattr(packages, package_name).__dict__.update(getattr(plugin, package_name).__dict__)

    return packages



from __future__ import annotations

import sys

import torch
from torch import nn

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class Mixup(DataAugmentationPyTorch):
    def __init__(self,
                 alpha: float = 1.0,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)

        self.alpha = alpha
        self.crossentropyloss = nn.CrossEntropyLoss()

        self.lambdadist: torch.distributions.beta.Beta | None = None

    def __mix_criterion(self,  # noqa: PLR0913
                        criterion: nn.Module,
                        pred: torch.Tensor,
                        y_a: torch.Tensor,
                        y_b: torch.Tensor,
                        l: torch.Tensor) -> torch.Tensor:
        l = l.mean() # same for whole batch
        return l * criterion(pred, y_a) + (1 - l) * criterion(pred, y_b)

    def loss(self, pred: torch.Tensor, targets: tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> torch.Tensor:
        return self.__mix_criterion(self.crossentropyloss, pred, *targets)

    def apply(self,
              x: torch.Tensor,
              y: torch.Tensor,
              device: torch.device) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """Return mixed inputs, pairs of targets, and lambda."""
        if self.lambdadist is None: # Lazy init so that CUDA is only used when module is actually called
            self.lambdadist = torch.distributions.beta.Beta(torch.tensor(self.alpha, device=device),
                                                            torch.tensor(self.alpha, device=device))

        l = self.lambdadist.sample()

        batch_size = x.shape[0]
        index = torch.randperm(batch_size)

        mixed_x = l * x + (1 - l) * x[index, :]
        y_a, y_b = y, y[index]

        return mixed_x, (y_a, y_b, l)

    @override
    def __call__(self,
                 data: tuple[torch.Tensor, torch.Tensor],
                 device: torch.device) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        return self.apply(*data, device=device)



import math
from collections import OrderedDict
from typing import Optional, Union

import numpy as np
import torch
from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity, QuantizedLinear, QuantizedReLU
from qualia_core.typing import RecursiveConfigDict
from torch import nn


class QuantizedCNN(nn.Sequential):
    def __init__(self,
                 input_shape: tuple[int],
                 output_shape: tuple[int],

                 filters: list[int] = [6, 16, 120],
                 kernel_sizes: list[int] = [3, 3, 5],
                 paddings: list[int] = [0, 0, 0],
                 strides: list[int] = [1, 1, 1],
                 batch_norm: bool = False,
                 dropouts: Union[float, list[float]] = 0.0,
                 pool_sizes: list[int] = [2, 2, 0],
                 fc_units: list[int] = [84],
                 prepool: Union[int, list[int]] = 1,
                 postpool: Union[int, list[int]]=1,

                 dims: int=1,

                 quantize_linear: bool = True,
                 fused_relu: bool = True,
                 quant_params: Optional[RecursiveConfigDict] = None) -> None:

        self.input_shape = input_shape
        self.output_shape = output_shape

        if quant_params and 'bits' in quant_params :
            quant_params['bits'] = int(quant_params['bits']) # Force conversion from TOML int to plain Python int
            if quant_params['bits'] < 1:
                raise ValueError('bits must be set to a strictly positive integer')
        else :
            raise ValueError('bits must exist in quant_params conf')

        if dims == 1:
            import qualia_core.learningmodel.pytorch.layers1d as layers_t
        elif dims == 2:
            import qualia_core.learningmodel.pytorch.layers2d as layers_t
        else:
            raise ValueError('Only dims=1 or dims=2 supported')

        # Backward compatibility for config not defining dropout as a list
        if not isinstance(dropouts, list):
            dropouts = [dropouts] * (len(filters) + len(fc_units))

        layers: OrderedDict[str, nn.Module] = OrderedDict()

        layers['identity1'] = QuantizedIdentity(quant_params=quant_params)

        if isinstance(prepool, int) and prepool > 1:
            layers['prepool'] = layers_t.QuantizedAvgPool(prepool, quant_params=quant_params)
        elif not isinstance(prepool, int) and math.prod(prepool) > 1:
            layers['prepool'] = layers_t.QuantizedAvgPool(tuple(prepool), quant_params=quant_params)

        layers['conv1'] = layers_t.QuantizedConv(in_channels=input_shape[-1],
                                                 out_channels=filters[0],
                                                 kernel_size=kernel_sizes[0],
                                                 padding=paddings[0],
                                                 stride=strides[0],
                                                 quant_params=quant_params,
                                                 activation=nn.ReLU() if fused_relu and not batch_norm else None)

        if batch_norm:
            layers['bn1'] = layers_t.QuantizedBatchNorm(filters[0],
                                                        quant_params=quant_params,
                                                        activation=nn.ReLU() if fused_relu else None)

        if not fused_relu:
            layers['relu1'] = QuantizedReLU(quant_params=quant_params)

        if dropouts[0]:
            layers['dropout1'] = nn.Dropout(dropouts[0])
        if pool_sizes[0]:
            layers['maxpool1'] = layers_t.QuantizedMaxPool(pool_sizes[0], quant_params=quant_params)

        i = 2
        for in_filters, out_filters, kernel, pool_size, padding, stride, dropout in zip(filters, filters[1:], kernel_sizes[1:],
                                                                                        pool_sizes[1:], paddings[1:], strides[1:],
                                                                                        dropouts[1:]):
            layers[f'conv{i}'] = layers_t.QuantizedConv(in_channels=in_filters,
                                                        out_channels=out_filters,
                                                        kernel_size=kernel,
                                                        padding=padding,
                                                        stride=stride,
                                                        quant_params=quant_params,
                                                        activation=nn.ReLU() if fused_relu and not batch_norm else None)

            if batch_norm:
                layers[f'bn{i}'] = layers_t.QuantizedBatchNorm(out_filters,
                                                               quant_params=quant_params,
                                                               activation=nn.ReLU() if fused_relu else None)

            if not fused_relu:
                layers[f'relu{i}'] = QuantizedReLU(quant_params=quant_params)

            if dropout:
                layers[f'dropout{i}'] = nn.Dropout(dropout)
            if pool_size:
                layers[f'maxpool{i}'] = layers_t.QuantizedMaxPool(pool_size, quant_params=quant_params)

            i += 1

        if isinstance(postpool, int) and postpool > 1:
            layers['postpool'] = layers_t.QuantizedAvgPool(postpool, quant_params=quant_params)
        if not isinstance(postpool, int) and math.prod(postpool) > 1:
            layers['postpool'] = layers_t.QuantizedAvgPool(tuple(postpool), quant_params=quant_params)

        layers['flatten'] = nn.Flatten()

        in_features = np.array(input_shape[:-1]) // np.array(prepool)
        for _, kernel, pool, padding, stride in zip(filters, kernel_sizes, pool_sizes, paddings, strides):
            in_features += np.array(padding) * 2
            in_features -= (kernel - 1)
            in_features = np.ceil(in_features / stride).astype(int)
            if pool:
                in_features = in_features // pool
        in_features = in_features // np.array(postpool)
        in_features = in_features.prod()
        in_features *= filters[-1]

        j = 1
        for in_units, out_units, dropout in zip((in_features, *fc_units), fc_units, dropouts[len(filters):]):
            if quantize_linear:
                layers[f'fc{j}'] = QuantizedLinear(in_units,
                                                   out_units,
                                                   quant_params=quant_params,
                                                   activation=nn.ReLU() if fused_relu else None)
            else:
                layers[f'fc{j}'] = nn.Linear(in_units, out_units)
            if not fused_relu or not quantize_linear:
                layers[f'relu{i}'] = QuantizedReLU(quant_params=quant_params)
            if dropout:
                layers[f'dropout{i}'] = nn.Dropout(dropout)

            i += 1
            j += 1

        if quantize_linear:
            layers[f'fc{j}'] = QuantizedLinear(fc_units[-1] if len(fc_units) > 0 else in_features,
                                               output_shape[0],
                                               quant_params=quant_params)
        else:
            layers[f'fc{j}'] = nn.Linear(fc_units[-1] if len(fc_units) > 0 else in_features, output_shape[0])

        super().__init__(layers)



def MLP(input_shape, output_shape, units: tuple=(100, 100, 100)):
    from tensorflow.keras.layers import Input, Dense, Flatten, Activation 
    from tensorflow.keras.models import Sequential

    model = Sequential()
    model.add(Input(input_shape))
    model.add(Flatten())
    for unit in units:
        model.add(Dense(unit))
        model.add(Activation('relu'))

    model.add(Dense(output_shape[0]))
    model.add(Activation('softmax'))
    
    return model



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic

from qualia_core.learningframework.LearningFramework import LearningFramework, T
from qualia_core.typing import TYPE_CHECKING, ModelConfigDict

if TYPE_CHECKING:
    from qualia_core.qualia import TrainResult  # noqa: TCH001

class PostProcessing(ABC, Generic[T]):
    @abstractmethod
    def __call__(self, trainresult: TrainResult, model_conf: ModelConfigDict) -> tuple[TrainResult, ModelConfigDict]:
        ...

    def process_name(self, name: str) -> str:
        return name

    def process_framework(self, framework: LearningFramework[T]) -> LearningFramework[Any]:
        return framework

    def process_mem_params(self, mem_params: int) -> Callable[[LearningFramework[T],
                                                               T], int]:
        def f(_: LearningFramework[T], __: T) -> int:
            return mem_params
        return f

    def process_model(self, model: T, model_conf: ModelConfigDict) -> tuple[T, ModelConfigDict]:
        return model, model_conf



from __future__ import annotations

import logging
import sys
from importlib.resources import files
from typing import Any

from qualia_core.deployment.toolchain import STM32CubeIDE
from qualia_core.evaluation.target.Qualia import Qualia as QualiaEvaluator
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.path import resources_to_path

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from qualia_core.postprocessing.Converter import Converter  # noqa: TCH001
    from qualia_core.postprocessing.QualiaCodeGen import QualiaCodeGen  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class NucleoL452REP(STM32CubeIDE):
    evaluator = QualiaEvaluator # Suggested evaluator

    def __init__(self) -> None:
        projectdir = resources_to_path(files('qualia_codegen_core.examples'))/'Qualia-CodeGen-NucleoL452REP'
        super().__init__(projectname='Qualia-CodeGen-NucleoL452REP',
                         projectdir=projectdir)

        self.__model_data = self._projectdir/'Core'/'Inc'/'model.h'

    def __write_model(self, model: QualiaCodeGen) -> None:
        if model.h is None:
            logger.error('Cannot write model source: QualiaCodeGen Converter did not run successfully (QualiaCodeGen.h is None)')
            raise ValueError

        with self.__model_data.open('w') as f:
            _ = f.write(model.h)

    @override
    def prepare(self,
                tag: str,
                model: Converter[Any],
                optimize: str,
                compression: int) -> Self | None:
        # Keep here for isinstance() to avoid circual import
        from qualia_core.postprocessing.QualiaCodeGen import QualiaCodeGen

        if optimize and optimize != 'cmsis-nn':
            logger.error('Optimization %s not available for %s', optimize, type(self).__name__)
            raise ValueError

        if compression != 1:
            logger.error('No compression available for %s', type(self).__name__)
            raise ValueError

        if not isinstance(model, QualiaCodeGen):
            logger.error('%s excepts the model to come from a QualiaCodeGen Converter', type(self).__name__)
            raise TypeError

        if model.directory is None:
            logger.error('QualiaCodeGen Converter did not run successfully (QualiaCodeGen.directory is None)')
            raise ValueError

        self._create_outdir()
        self._clean_workspace()
        self.__write_model(model=model)

        args: tuple[str, ...] = ('-include', str(model.directory.resolve()/'include'/'defines.h'))

        if optimize == 'cmsis-nn':
            args = (*args,
                    '-D', 'WITH_CMSIS_NN',
                    '-D', 'ARM_MATH_DSP')

        if not self._build(args=args):
            return None
        self._copy_buildproduct(tag=tag)
        return self



from pathlib import Path
import shutil
import subprocess
import sys
import os
from collections import namedtuple
from importlib.resources import files

from .TFLite2CArray import TFLite2CArray

class SparkFunEdge:
    import qualia_core.evaluation.target.Qualia as evaluator # Suggested evaluator

    def __init__(self,
        dev: str='/dev/ttyUSB0',
        baudrate: int=921600,
        projectdir: Path=files('qualia_core.assets')/'projects'/'tflitemicro'/'SparkFunEdge',
        tfpath: Path=Path('third_party/tflite-micro'),
        outdir: Path=Path('out/deploy/sparkfun_edge')):
        self.__dev = dev
        self.__projectdir = projectdir
        self.__projectdir_from_tfpath = Path(os.path.relpath(projectdir.resolve(), tfpath.resolve()))
        self.__tfpath = tfpath
        self.__baudrate = baudrate
        self.__outdir = outdir
        self.__numthreads = os.cpu_count()

    def __find_ambiqsuite(self, tfpath):
        return sorted((tfpath/'tensorflow'/'lite'/'micro'/'tools'/'make'/'downloads').glob('AmbiqSuite-*'), reverse=True)[0]

    def __create_outdir(self, outdir: Path):
        outdir.mkdir(parents=True, exist_ok=True)

    def __write_model(self, model, modelpath):
        with modelpath.open('w') as f:
            f.write(TFLite2CArray().convert(model.data, model.input_shape).cc)
    
    def __clean(self, tfpath):
        cmd = ['make', '-C', str(tfpath),
                         '-f', 'tensorflow/lite/micro/tools/make/Makefile',
                         '-f', str(self.__projectdir_from_tfpath/'Makefile.inc'),
                         'TARGET=sparkfun_edge',
                         'TARGET_ARCH=cortex-m4',
                         'clean']
        print(cmd)
        s = subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def __third_party_downloads(self, tfpath, cmsis):
        cmd = ['make', '-C', str(tfpath),
                         '-j', '1',
                         '-f', 'tensorflow/lite/micro/tools/make/Makefile',
                         '-f', str(self.__projectdir_from_tfpath/'Makefile.inc'),
                         'TARGET=sparkfun_edge',
                         'TARGET_ARCH=cortex-m4',
                         'third_party_downloads']
        if cmsis:
            cmd.append('OPTIMIZED_KERNEL_DIR=cmsis_nn')
        print(cmd)
        s = subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def __build(self, tfpath, cmsis):
        cmd = ['make', '-C', str(tfpath),
                         '-j', str(self.__numthreads),
                         '-f', 'tensorflow/lite/micro/tools/make/Makefile',
                         '-f', str(self.__projectdir_from_tfpath/'Makefile.inc'),
                         'TARGET=sparkfun_edge',
                         'TARGET_ARCH=cortex-m4',
                         'myapp_bin']
        if cmsis:
            cmd.append('OPTIMIZED_KERNEL_DIR=cmsis_nn')
        print(cmd)
        s = subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def __sign(self, tfpath, outdir, tag: str):
        # use default  keys
        d = self.__find_ambiqsuite(self.__tfpath)/'tools'/'apollo3_scripts'
        shutil.copy(d/'keys_info0.py', d/'keys_info.py')

        s = subprocess.run(['python3',
            str(self.__find_ambiqsuite(self.__tfpath)/'tools'/'apollo3_scripts'/'create_cust_image_blob.py'),
            '--bin', str(tfpath/'tensorflow'/'lite'/'micro'/'tools'/'make'/'gen'/'sparkfun_edge_cortex-m4_default'/'bin'/'myapp.bin'),
            '--load-address', '0xC000', '--magic-num', '0xCB', '-o', str(outdir/f'{tag}_nonsecure_ota'), '--version', '0x0'],
            stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def __create_image(self, tfpath, outdir, tag: str):
        s = subprocess.run(['python3',
            str(self.__find_ambiqsuite(self.__tfpath)/'tools'/'apollo3_scripts'/'create_cust_wireupdate_blob.py'),
            '--load-address', '0x20000', '--bin', str(outdir/f'{tag}_nonsecure_ota.bin'), '-i', '6', '-o', str(outdir/f'{tag}_nonsecure_wire'),
            '--options', '0x1'],
            stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def __upload(self, dev: str, baudrate: int, tfpath, outdir, tag: str):
        s = subprocess.run(['python3',
            str(self.__find_ambiqsuite(self.__tfpath)/'tools'/'apollo3_scripts'/'uart_wired_update.py'),
            '-b', str(baudrate), dev, '-r', '1', '-f', str(outdir/f'{tag}_nonsecure_wire.bin'), '-i', '6'],
            stdout=sys.stdout, stderr=sys.stderr, text=True)
        return s.returncode == 0

    def prepare(self, tag, model, optimize=None, compression: int=None):
        if compression != 1:
            raise ValueError(f'Compression not supported by {self.__class__.__name__}')

        if not optimize:
            cmsis = False
        elif optimize == 'cmsis-nn':
            cmsis = True # Use CMSIS-NN kernel in TFLiteMicro
            print(f'FIXME: Evaluation not logging optimization')
        else:
            raise ValueError(f'Unsupported {optimize} optimization by {self.__class__.__name__}')

        self.__create_outdir(outdir=self.__outdir)

        if not self.__clean(tfpath=self.__tfpath):
            return None
        self.__write_model(model, modelpath=self.__projectdir/'model_data.cc')
        if not self.__third_party_downloads(tfpath=self.__tfpath, cmsis=cmsis):
                return None
        if not self.__build(tfpath=self.__tfpath, cmsis=cmsis):
            return None
        if not self.__sign(tfpath=self.__tfpath, outdir=self.__outdir, tag=tag):
            return None
        if not self.__create_image(tfpath=self.__tfpath, outdir=self.__outdir, tag=tag):
            return None
        return self

    def deploy(self, tag):
        input('Put target in programming mode and press Enter…')
        if not self.__upload(dev=self.__dev, baudrate=self.__baudrate, tfpath=self.__tfpath, outdir=self.__outdir, tag=tag):
            return None
        return namedtuple('Deploy', ['rom_size', 'ram_size', 'evaluator'])(self.__rom_size(tag), None, self.evaluator)

    def __rom_size(self, tag):
        return (self.__outdir/f'{tag}_nonsecure_wire.bin').stat().st_size




class ResBlock: # Not an actual Keras model because it is harder to serialize and get it to work on third-party tools
    expansion = 1

    def __init__(self,
                layers_t,
                in_filters,
                filters,
                kernel_size,
                stride,
                padding,
                batch_norm: bool=False,
                bn_momentum: float=0.99,
                force_projection_with_stride: bool=True):
        super().__init__()
        from tensorflow.keras.layers import BatchNormalization, Activation, Add
        from tensorflow.python.keras.backend import unique_object_name

        self.in_filters = in_filters
        self.filters = filters
        self.stride = stride
        self.batch_norm = batch_norm
        self.force_projection_with_stride = force_projection_with_stride
        
        self.padding_1 = layers_t.ZeroPadding(padding=padding)
        self.conv_1 = layers_t.Conv(filters=filters, kernel_size=kernel_size, strides=1, use_bias=not batch_norm)
        if batch_norm:
            self.batchnorm_1 = BatchNormalization(momentum=bn_momentum)
        if self.stride != 1:
            self.maxpool_1 = layers_t.MaxPooling(stride)
        self.relu_1 = Activation('relu')

        self.padding_2 = layers_t.ZeroPadding(padding=padding)
        self.conv_2 = layers_t.Conv(name=unique_object_name('conv_ref', zero_based=True), filters=filters, kernel_size=kernel_size, strides=1, use_bias=not batch_norm)
        if batch_norm:
            self.batchnorm_2 = BatchNormalization(momentum=bn_momentum)

        if self.stride != 1:
            self.maxpool_shortcut = layers_t.MaxPooling(stride)

        if self.in_filters != self.expansion*self.filters or force_projection_with_stride and self.stride != 1:
            self.conv_shortcut = layers_t.Conv(name=unique_object_name('conv_shortcut', zero_based=True), filters=self.expansion * filters, kernel_size=1, strides=1, use_bias=not batch_norm)
            if batch_norm:
                self.batchnorm_shortcut = BatchNormalization(momentum=bn_momentum)

        self.add = Add()
        self.relu_2 = Activation('relu')


    def __call__(self, inputs):
        x = self.padding_1(inputs)
        x = self.conv_1(x)
        if self.batch_norm:
            x = self.batchnorm_1(x)
        if self.stride != 1:
            x = self.maxpool_1(x)
        x = self.relu_1(x)
        
        x = self.padding_2(x)
        x = self.conv_2(x)
        if self.batch_norm:
            x = self.batchnorm_2(x)

        # shortcut
        tmp = inputs
        if self.in_filters != self.expansion*self.filters or self.force_projection_with_stride and self.stride != 1:
            tmp = self.conv_shortcut(tmp)
            if self.batch_norm:
                tmp = self.batchnorm_shortcut(tmp)
        if self.stride != 1:
            tmp = self.maxpool_shortcut(tmp)

        x = self.add([x, tmp])
        x = self.relu_2(x)

        return x

    @property
    def output_shape(self):
        return self.relu_2.output_shape

def ResNet(input_shape,
        output_shape,
        filters: int=(15, 15, 30, 60, 120), # + (120,)
        kernel_sizes: tuple=(3, 3, 3, 3, 3),

        num_blocks: tuple=(2, 2, 2, 2), # + (2,)
        strides: tuple=(1, 1, 2, 2, 2), # + (2,)
        paddings: int=(1, 1, 1, 1, 1),

        prepool: int=1,
        batch_norm: bool=False,
        bn_momentum: float=0.99,
        force_projection_with_stride: bool=True,

        dims: int=1):

    import tensorflow as tf
    from tensorflow.keras import Model
    from tensorflow.keras.models import Sequential
    from tensorflow.keras.layers import Input, Flatten, Dense, BatchNormalization, Activation
    import tensorflow.keras.backend as backend

    if dims == 1:
        import qualia_core.learningmodel.keras.layers1d as layers_t
    elif dims == 2:
        import qualia_core.learningmodel.keras.layers2d as layers_t
    else:
        raise ValueError('Only dims=1 or dims=2 supported')

    # Do not track moving statistics (avg/var) for inference, use batch statistics only like PyTorch's track_running_stats
    track_running_stats ={} # {'training': True}

    # Actual model definition
    # Conv → BatchNorm → ResBlock → ResBlock → ResBlock → AveragePooling → Flatten → FullyConnected(out)

    inputs = Input(shape=input_shape)

    if prepool > 1:
        x = layers_t.AveragePooling(pool_size=prepool)(inputs)
    else:
        x = inputs

    # First non-residual convolution
    x = layers_t.ZeroPadding(padding=paddings[0])(x)
    x = layers_t.Conv(filters=filters[0], kernel_size=kernel_sizes[0], strides=strides[0], use_bias=not batch_norm)(x)
    if batch_norm:
        x = BatchNormalization(momentum=bn_momentum)(x)
    x = Activation('relu')(x)

    # Residual layers
    in_filters = filters[0]
    for f, kernel_size, layer_stride, padding, num_blocks in zip(filters[1:], kernel_sizes[1:], strides[1:], paddings[1:], num_blocks):
        for block_stride in [layer_stride] + [1] * (num_blocks - 1):
            block = ResBlock(layers_t=layers_t,
                        in_filters=in_filters,
                        filters=f,
                        kernel_size=kernel_size,
                        stride=block_stride,
                        padding=padding,
                        batch_norm=batch_norm,
                        bn_momentum=bn_momentum) # 1st block of layer
            x = block(x)
            in_filters = f * block.expansion

    # Reduce last conv layer output feature maps
    x = layers_t.MaxPooling(x.shape[1])(x)
    x = Flatten()(x)
    x = Dense(output_shape[0], use_bias=True)(x)
    outputs = Activation('softmax')(x)

    model = Model(inputs=inputs, outputs=outputs, name='ResNet')

    return model



import sys

import torch
import torch.nn.quantized
from torch import nn

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class Add(nn.Module):
    def __init__(self) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()
        self.functional = torch.nn.quantized.FloatFunctional()

    @override
    def forward(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.functional.add(a, b)

    @override
    def __call__(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return super().__call__(a, b)



# See https://www.tensorflow.org/lite/performance/post_training_quantization

import functools
import importlib

class Keras2TFLite:
    # TFLite can be deployed both with STM32CubeAI on STM32 boards and TFLiteMicro on other boards (SparkFun Edge…)
    import qualia_core.deployment.tflitemicro as deployers # Use TFLiteMicro deployers
    deployers.__dict__.update(importlib.import_module('qualia_core.deployment.stm32cubeai').__dict__) # Merge STM32CubeAI deployers

    def __init__(self, quantize: int=-1, new_converter=False):
        if quantize not in ['float32', 'float16', 'float8', 'int8', 'int16']:
            raise ValueError(self.__class__.__name__ + ' only suports none, 8 or 16-bit quantization')
        self.__quantize = quantize
        self.__new_converter = new_converter

        if quantize == 'float32':
            self.__width = 32
        elif quantize == 'float16' or quantize == 'int16':
            self.__width = 16
        elif quantize == 'float8' or quantize == 'int8':
            self.__width = 8

    def representative_dataset_gen(self, representative_dataset):
        for input_value in representative_dataset:
            yield [input_value.reshape([1, *input_value.shape])]

    def convert(self, framework, keras_model, model_name, representative_dataset=None):
        import tensorflow as tf
        converter = tf.lite.TFLiteConverter.from_keras_model(keras_model)
        converter.experimental_new_converter = self.__new_converter

        if self.__quantize == 'float16':
            raise ValueError('float16 quantization not supported by TFLite Micro')
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
            converter.target_spec.supported_types = [tf.float16]
        elif self.__quantize == 'float8':
            raise ValueError('float8 quantization not supported by TFLite Micro')
            converter.optimizations = [tf.lite.Optimize.OPTIMIZE_FOR_SIZE]
        elif self.__quantize == 'int8':
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
            converter.representative_dataset = functools.partial(self.representative_dataset_gen, representative_dataset)
            converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
            # We use float inputs/outputs
            #converter.inference_input_type = tf.int8
            #converter.inference_output_type = tf.int8
            converter.inference_input_type = tf.float32
            converter.inference_output_type = tf.float32
        elif self.__quantize == 'int16':
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
            converter.representative_dataset = functools.partial(self.representative_dataset_gen, representative_dataset)
            converter.target_spec.supported_ops = [tf.lite.OpsSet.EXPERIMENTAL_TFLITE_BUILTINS_ACTIVATIONS_INT16_WEIGHTS_INT8]
            #converter.inference_input_type = tf.int16
            #converter.inference_output_type = tf.int16
            # Old converter is buggy for int8/int16 quant, it cannot use float inputs or representative_dataset
            converter.experimental_new_converter = True
        elif self.__quantize != 'float32':
            raise NotImplemented(f'{self.__quantize} no implemented in {self.__class__.__name__}')

        self.__data = converter.convert()
        self.__input_shape = keras_model.input_shape

        return self

    @property
    def data(self) -> bytes:
        return self.__data

    @property
    def input_shape(self):
        return self.__input_shape

    def process_mem_params(self, mem_params):
        return lambda framework, model: (framework.n_params(model) * self.__width) // 8



from __future__ import annotations

from typing import Protocol

from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qualia_core.learningmodel.pytorch.Quantizer import Quantizer  # noqa: TCH001

class QuantizerInputProtocol(Protocol):
    def __init__(self, *_: object, **__: object) -> None:
        super().__init__()

    quantizer_input: Quantizer

class QuantizerActProtocol(Protocol):
    def __init__(self, *_: object, **__: object) -> None:
        super().__init__()

    quantizer_act: Quantizer

class QuantizerWProtocol(Protocol):
    def __init__(self, *_: object, **__: object) -> None:
        super().__init__()

    quantizer_w: Quantizer

class QuantizedLayer:
    def __init__(self, *_: object, **__: object) -> None:
        super().__init__()
        self.quantizer_input: Quantizer | None = None
        self.quantizer_act: Quantizer | None = None
        self.quantizer_w: Quantizer | None = None

    @property
    def input_q(self) -> int | None:
        """Number of bits used to encode the fractional part of the input in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the input or ``None`` if not applicable.
        """
        if self.quantizer_input is None:
            return None
        return self.quantizer_input.fractional_bits

    @property
    def activation_q(self) -> int | None:
        """Number of bits used to encode the fractional part of the output in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the output or ``None`` if not applicable.
        """
        if self.quantizer_act is None:
            return None
        return self.quantizer_act.fractional_bits

    @property
    def weights_q(self) -> int | None:
        """Number of fractional part bits for the weights in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the weights or ``None`` if not applicable.
        """
        if self.quantizer_w is None:
            return None
        return self.quantizer_w.fractional_bits

    @property
    def input_round_mode(self) -> str | None:
        if self.quantizer_input is None:
            return None
        return self.quantizer_input.roundtype

    @property
    def activation_round_mode(self) -> str | None:
        if self.quantizer_act is None:
            return None
        return self.quantizer_act.roundtype

    @property
    def weights_round_mode(self) -> str | None:
        if self.quantizer_w is None:
            return None
        return self.quantizer_w.roundtype



from __future__ import annotations

import logging
import sys
import time
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import numpy.typing

from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets
from qualia_core.utils.file import CSVReader, DirectoryReader

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class GTSRB(RawDataset):
    @dataclass
    class Row:
        Filename: str
        Width: str
        Height: str
        Roi_X1: str
        Roi_Y1: str
        Roi_X2: str
        Roi_Y2: str
        ClassId: str

    def __init__(self,
                 path: str='',
                 width: int=32,
                 height: int=32) -> None:
        super().__init__()
        self.__dr = DirectoryReader()
        self.__csvreader: CSVReader[GTSRB.Row] = CSVReader()
        self.__path = Path(path)
        self.__width = width
        self.__height = height
        self.sets.remove('valid')

    def __load_train(self, path: Path) -> RawData:
        import imageio
        from skimage import transform
        start = time.time()
        train_dir = self.__dr.read(path/'Final_Training'/'Images', ext='.ppm', recursive=True)

        train_x_list: list[numpy.typing.NDArray[np.float32]] = []
        train_y_list: list[int] = []
        for image in train_dir:
            imdata = imageio.imread(image)
            imdata = transform.resize(imdata, (self.__width, self.__height))
            imdata = imdata.astype(np.float32)
            train_x_list.append(imdata)
            train_y_list.append(int(image.parent.name))
        train_x = np.array(train_x_list)
        train_y = np.array(train_y_list)

        logger.info('__load_train(): Elapsed: %s s', time.time() - start)

        return RawData(train_x, train_y)

    def __load_test(self, path: Path) -> RawData:
        import imageio
        from skimage import transform
        start = time.time()
        test_dir = path/'Final_Test'/'Images'
        gt = self.__csvreader.read(path/'GT-final_test.csv',
                                   delimiter=';',
                                   labels=GTSRB.Row,
                                   skip_header=True)
        test_x_list: list[numpy.typing.NDArray[np.float32]] = []
        test_y_list: list[int] = []
        for img in gt.content:
            imdata = imageio.imread(test_dir/(img.Filename))
            imdata = transform.resize(imdata, (self.__width, self.__height))
            imdata = imdata.astype(np.float32)
            test_x_list.append(imdata)
            test_y_list.append(int(img.ClassId))
        test_x = np.array(test_x_list)
        test_y = np.array(test_y_list)

        logger.info('__load_test(): Elapsed: %s s', time.time() - start)

        return RawData(test_x, test_y)

    @override
    def __call__(self) -> RawDataModel:
        return RawDataModel(sets=RawDataSets(train=self.__load_train(self.__path), test=self.__load_test(self.__path)),
                            name=self.name)



from .DataModel import DataModel
from .FeatureSample import FeatureSample
from .RawDataModel import RawDataModel
from .TimeSample import TimeSample
from .WindowedSample import WindowedSample

__all__ = ['DataModel', 'FeatureSample', 'RawDataModel', 'TimeSample', 'WindowedSample']



from __future__ import annotations

import logging
import sys

import torch

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class Rotation(DataAugmentationPyTorch):

    def __init__(self,  # noqa: PLR0913
                 dimensions: int=3,
                 sigma: float=0.375,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)

        if dimensions != 3:  # noqa: PLR2004
            logger.error('Only 3D rotations are supported, dimensions=%s', dimensions)
            raise ValueError
        self.dimensions = dimensions
        self.sigma = sigma

        self.angledist = None

    def apply(self, x: torch.Tensor, y: torch.Tensor, device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        import pytorch3d.transforms

        if self.angledist is None: # Lazy init so that CUDA is only used when module is actually called
            self.angledist = torch.distributions.normal.Normal(torch.tensor(0.0, device=device),
                                                               torch.tensor(self.sigma, device=device))

        angle = self.angledist.sample((3, ))

        rotmat = pytorch3d.transforms.axis_angle_to_matrix(angle)

        transform = pytorch3d.transforms.Rotate(rotmat, device=device)

        # For some reason pytorch3d transforms work with 3D channels last
        x = torch.swapaxes(x, 1, 2)

        rotated_x_array = tuple(transform.transform_points(x[:, :, i:i+self.dimensions])
                                    if x.shape[-1] >= i + self.dimensions else x[:, :, i:]
                                    for i in range(0, x.shape[-1], self.dimensions))
        rotated_x = torch.cat(rotated_x_array, dim=-1)

        rotated_x = torch.swapaxes(rotated_x, 1, 2)

        return rotated_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data, device=device)



from .BandPassFilter import BandPassFilter
from .Class2BinMatrix import Class2BinMatrix
from .DatamodelConverter import DatamodelConverter
from .DatasetSplitter import DatasetSplitter
from .DatasetSplitterBySubjects import DatasetSplitterBySubjects
from .MFCC import MFCC
from .Normalize import Normalize
from .PrintHARDataModelSummary import PrintHARDataModelSummary
from .RemoveActivity import RemoveActivity
from .RemoveSensor import RemoveSensor
from .Reshape2DTo1D import Reshape2DTo1D
from .VisualizeActivities import VisualizeActivities
from .VisualizeWindows import VisualizeWindows
from .Window import Window

__all__ = ['BandPassFilter',
           'Class2BinMatrix',
           'DatamodelConverter',
           'DatasetSplitterBySubjects',
           'DatasetSplitter',
           'MFCC',
           'Normalize',
           'PrintHARDataModelSummary',
           'RemoveActivity',
           'RemoveActivity',
           'RemoveSensor',
           'Reshape2DTo1D',
           'VisualizeActivities',
           'VisualizeWindows',
           'Window']



from pathlib import Path
from collections import namedtuple

class Linux:
    import qualia_core.evaluation.host.TFLite as evaluator # Suggested evaluator

    def __create_outdir(self, modeloutdir):
        modeloutdir.mkdir(parents=True, exist_ok=True)

    def __write_model(self, model, modelpath):
        with modelpath.open('wb') as f:
            f.write(model.data)

    def prepare(self, tag, model, optimize: str, compression: int):
        modeloutdir = Path('out')/'deploy'/'tflite'
        modelpath = modeloutdir/f'{tag}.tflite'
        self.__create_outdir(modeloutdir)
        self.__write_model(model=model, modelpath=modelpath)

        return self

    def deploy(self, tag):
        print(self.__class__.__name__, 'Info: running locally, nothing to deploy')

        return namedtuple('Deploy', ['rom_size', 'evaluator'])(self.__rom_size(tag), self.evaluator)

    def __rom_size(self, tag):
        return -1



from __future__ import annotations

from abc import abstractmethod

from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking
from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pytorch_lightning.loggers import Logger  # noqa: TCH002

class ExperimentTrackingPyTorch(ExperimentTracking):
    @property
    @abstractmethod
    def logger(self) -> Logger | None:
        ...



import logging
import sys
from pathlib import Path
from typing import Final

import numpy as np
import numpy.typing

from qualia_core.datamodel import FeatureSample, WindowedSample
from qualia_core.datamodel.har import Activities, Activity, HARDataModel, Subject

from .HARDataset import HARDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class UCI_HAR(HARDataset):
    activitylist: Final[dict[str, Activities]] = {
            '1': Activities.WALKING,
            '2': Activities.WALKING_UPSTAIRS,
            '3': Activities.WALKING_DOWNSTAIRS,
            '4': Activities.SITTING,
            '5': Activities.STANDING,
            '6': Activities.LYING,
            }

    def __init__(self, path: str='', variant: str='features') -> None:
        super().__init__()
        self.__path = Path(path)
        self.__variant = variant
        self.sets.remove('valid')

    def __load_features_data(self, part: str, path: Path) -> list[Subject]:
        # load features, labels and subjects
        with (path/part/f'X_{part}.txt').open('r') as fxtrain, \
             (path/part/f'y_{part}.txt').open('r') as fytrain, \
             (path/part/f'subject_{part}.txt').open('r') as fsubject:

             sxtrain = fxtrain.read().splitlines()
             sytrain = fytrain.read().splitlines()
             ssubject = fsubject.read().splitlines()

        # populate datamodel
        subjects: list[Subject] = []
        activities: list[Activity] = []
        samples: list[FeatureSample] = []
        lastact = sytrain[0]
        lastsub = ssubject[0]

        for data, label, subject in zip(sxtrain, sytrain, ssubject):
            if lastact != label or lastsub != subject:
                activities.append(Activity(UCI_HAR.activitylist[lastact], samples))
                samples = []
                lastact = label

            if lastsub != subject:
                subjects.append(Subject(subject, activities))
                activities = []
                lastsub = subject

            samples.append(FeatureSample([float(d) for d in data.split()]))

        activities.append(Activity(UCI_HAR.activitylist[label], samples))

        subjects.append(Subject(subject, activities, part=part))

        return subjects

    def __load_features(self, path: Path) -> HARDataModel:
        return HARDataModel(sets=HARDataModel.Sets(self.__load_features_data('train', path=path),
                                                   self.__load_features_data('test', path=path)),
                            name=self.name)

    def __load_raw_data(self, part: str, path: Path) -> list[Subject]:
        filenames = [f'{sensorname}_{dim}_{part}.txt'
                        for sensorname in ('body_acc', 'body_gyro', 'total_acc')
                            for dim in ('x', 'y', 'z')]

        # Load body_acc,body_gyro,total_acc x,y,z data and concatenate to create (,128,7)
        xtrain_list: list[numpy.typing.NDArray[np.float32]] = []
        for fn in filenames:
            with (path/part/'Inertial Signals'/fn).open() as f:
                xtrain_list.append(np.array([[float(d) for d in line.rstrip().split()] for line in f.readlines()],
                                             dtype=np.float32))
        xtrain = np.dstack(xtrain_list)

        # load labels and subjects
        with (path/part/f'y_{part}.txt').open('r') as fytrain, \
             (path/part/f'subject_{part}.txt').open('r') as fsubject:

            sytrain = fytrain.read().splitlines()
            ssubject = fsubject.read().splitlines()

        # populate datamodel
        subjects: list[Subject] = []
        activities: list[Activity] = []
        samples: list[WindowedSample] = []
        lastact = sytrain[0]
        lastsub = ssubject[0]


        for data, label, subject in zip(xtrain, sytrain, ssubject):
            if lastact != label or lastsub != subject:
                activities.append(Activity(int(float(lastact))-1, samples))
                samples = []
                lastact = label

            if lastsub != subject:
                subjects.append(Subject(lastsub, activities))
                activities = []
                lastsub = subject
            samples.append(WindowedSample(data))

        activities.append(Activity(int(float(lastact))-1, samples))

        subjects.append(Subject(lastsub, activities, part=part))

        return subjects

    def __load_raw(self, path: Path) -> HARDataModel:
        return HARDataModel(sets=HARDataModel.Sets(train=self.__load_raw_data('train', path=path),
                                                   test=self.__load_raw_data('test', path=path)),
                            name=self.name)

    @override
    def __call__(self) -> HARDataModel:
        if self.__variant == 'features':
            return self.__load_features(self.__path)

        if self.__variant == 'raw':
            return self.__load_raw(self.__path)

        logger.error('Only features and raw variants supported')
        raise ValueError

    @property
    @override
    def name(self) -> str:
        return f'{super().name}_{self.__variant}'



"""Holds a shared instance of a numpy Generator with set seed."""
import numpy as np


class GeneratorContainer:
    """Simple container for a numpy Generator. Default to random seed before seed() is called."""

    __generator: np.random.Generator = np.random.default_rng()

    def seed(self, seed: int) -> None:
        """Replace the contained numpy Generator with a new generator with set seed."""
        self.__generator = np.random.default_rng(seed=seed)

    @property
    def generator(self) -> np.random.Generator:
        """Returns the container numpy Generator."""
        return self.__generator

shared = GeneratorContainer()

__all__ = ['shared']



from __future__ import annotations

import logging
from dataclasses import dataclass, fields
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

logger = logging.getLogger(__name__)

T = TypeVar('T')
U = TypeVar('U')

class DataModel(Generic[T]):
    sets: DataModel.Sets[T]
    name: str

    @dataclass
    class Sets(Generic[U]):
        train: U | None = None
        valid: U | None = None
        test: U | None = None

        @classmethod
        def fieldnames(cls) -> tuple[str, ...]:
            return tuple(f.name for f in fields(cls))

        def asdict(self) -> dict[str, U]:
            # Explicitely use vars rather than dataclasses.asdict() since we don't want to recurse nor copy
            return vars(self)

        def __iter__(self) -> Iterator[tuple[str, U]]:
            # Skip non-existent sets
            return {k: v for k,v in self.asdict().items() if v is not None}.items().__iter__()

        def export(self, path: Path) -> None:
            for sname, sdata in self.asdict().items():
                if not sdata:
                    logger.info('No "%s" set to export', sname)
                else:
                    (path/sname).mkdir(parents=True, exist_ok=True)
                    sdata.export(path/sname)

    def __init__(self, sets: Sets[T], name: str) -> None:
        super().__init__()
        self.sets = sets
        self.name = name

    def __iter__(self) -> Iterator[tuple[str, T]]:
        return self.sets.__iter__()

    def export(self) -> None:
        self.sets.export(Path('out')/'data'/self.name)



from .Preprocessing import Preprocessing

from qualia_core.utils.logger import TextLogger

class PrintHARDataModelSummary(Preprocessing):
    log = TextLogger(name=__name__)

    def __print_subjects_summary(self, name, s, activitylabels):
        msg = f'Set: {name}\n'
        msg += '      '
        for act in activitylabels:
            msg += f'  {act.name}'
        msg += '  TOTAL\n'

        ss = sorted(s, key=lambda s: int(s.name[1:])) # Sort by subject number (without T prefix)

        total_a_count = [0] * len(activitylabels)

        for subject in ss:
            msg += ' ' * (5 - len(subject.name))
            msg += f'{subject.name}: '
            a_count = [0] * len(activitylabels)
            for a in subject.activities:
                a_count[activitylabels.index(a.kind)] += len(a.get_raw_array())
                total_a_count[activitylabels.index(a.kind)] += len(a.get_raw_array())

            for a, c in enumerate(a_count):
                msg += ' ' * (len(activitylabels[a].name) - len(str(c)) + 1)
                msg += f'{c} '
            total = sum(a_count)
            msg += ' ' * (len('TOTAL') - len(str(total)) + 1)
            msg += f'{total}\n'

        # Total of activities across all subjects
        msg += 'Total: '

        for a, c in enumerate(total_a_count):
            msg += ' ' * (len(activitylabels[a].name) - len(str(c)) + 1)
            msg += f'{c} '
        total = sum(total_a_count)
        msg += ' ' * (len('TOTAL') - len(str(total)) + 1)
        msg += f'{total}\n'

        print(msg, end='')
        self.log(msg)

        return total_a_count

    def __print_set_summary(self, sets_total, activitylabels):
        msg = '      '
        for act in activitylabels:
            msg += f'  {act.name}'
        msg += '  TOTAL\n'

        global_a_total = [sum([s[i] for s in sets_total.values()]) for i, _ in enumerate(activitylabels)]
        global_total = sum(global_a_total)
        sets_percentage = {name: [s[i] / global_a_total[i] for i, _ in enumerate(activitylabels)] for name, s in sets_total.items()}

        for (name, s), stot in zip(sets_percentage.items(), sets_total.values()):
            msg += ' ' * (5 - len(name))
            msg += f'{name}: '

            for a, c in enumerate(s):
                c = int(c * 100)
                msg += ' ' * (len(activitylabels[a].name) - len(str(c)))
                msg += f'{c}% '

            total = int(sum(stot) * 100 / global_total)
            msg += ' ' * (len('TOTAL') - len(str(total)))
            msg += f'{total}%\n'

        print(msg, end='')
        self.log(msg)

    def __call__(self, datamodel):
        # First generate the class numbers according to the activities present in the dataset
        # Duplicated from DatamodelConverter since they are not generated yet
        activitylabels = set()
        for sname, s in datamodel.sets:
            for subject in s:
                for activity in subject.activities:
                    activitylabels.add(activity.kind)
        activitylabels = sorted(activitylabels)

        sets_total = {}
        for name, s in datamodel:
            sets_total[name] = self.__print_subjects_summary(name, s, activitylabels)
        self.__print_set_summary(sets_total, activitylabels)

        return datamodel



from __future__ import annotations

import copy
import dataclasses
import logging
import sys
from collections.abc import Generator
from pathlib import Path
from typing import Any, Callable, NamedTuple, cast

from torch import nn

from qualia_core.learningframework.PyTorch import PyTorch
from qualia_core.typing import TYPE_CHECKING, ModelConfigDict, ModelParamsConfigDict, OptimizerConfigDict, RecursiveConfigDict
from qualia_core.utils.logger import CSVLogger
from qualia_core.utils.merge_dict import merge_dict

from .PostProcessing import PostProcessing

if TYPE_CHECKING:
    import keras  # type: ignore[import-untyped]  # noqa: TCH002
    import numpy.typing  # noqa: TCH002

    from qualia_core.datamodel.RawDataModel import RawData  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework, T  # noqa: TCH001
    from qualia_core.qualia import TrainResult  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class QuantizationAwareTrainingLoggerFields(NamedTuple):
    i: int
    name: str
    step: str
    width: int
    metrics: dict[str, int | float | numpy.typing.NDArray[Any]]

class QuantizationAwareTraining(PostProcessing[nn.Module]):

    def __init__(self,  # noqa: PLR0913
                 epochs: int = 1,
                 batch_size: int = 32,
                 model: RecursiveConfigDict | None = None,
                 optimizer: OptimizerConfigDict | None = None,
                 evaluate_before: bool = True) -> None:  # noqa: FBT002, FBT001
        super().__init__()

        self.qmodel = model if model is not None else {}
        self.epochs = epochs
        self.batch_size = batch_size
        self.qoptimizer = optimizer
        self._evaluate_before = evaluate_before

        if 'params' not in self.qmodel or not isinstance(self.qmodel['params'], dict):
            logger.error("'model.params' is required to be a table")
            raise ValueError
        if 'quant_params' not in self.qmodel['params'] or not isinstance(self.qmodel['params']['quant_params'], dict):
            logger.error("'model.params.quant_params' is required to be a table")
            raise ValueError
        self.qmodel_params: ModelParamsConfigDict = self.qmodel['params'] # TODO: validate

        quant_params = self.qmodel_params['quant_params']

        self.width: int = quant_params.get('bits', 0)
        self.force_q: int | None = quant_params.get('force_q', None)
        self.lsq: bool = quant_params.get('LSQ', False)

        self.log: CSVLogger[QuantizationAwareTrainingLoggerFields] = CSVLogger(name=__name__)
        self.log.fields = QuantizationAwareTrainingLoggerFields

    def __evaluate(self,  # noqa: PLR0913
                   framework: LearningFramework[T],
                   model: T,
                   trainset: RawData,
                   testset: RawData,
                   trainresult: TrainResult,
                   step: str) -> tuple[dict[str, int | float | numpy.typing.NDArray[Any]], str]:
        logger.info('Evaluation on train dataset')
        name = f'{trainresult.name}_q{self.width}_r{trainresult.i}_{step}_eval_train'
        metrics = framework.evaluate(model,
                                     trainset,
                                     batch_size=self.batch_size,
                                     dataaugmentations=trainresult.dataaugmentations,
                                     experimenttracking=trainresult.experimenttracking,
                                     dataset_type='train',
                                     name=name)

        if len(testset.x) > 0: # Don't evaluate if testset is empty
            logger.info('Evaluation on test dataset')
            name = f'{trainresult.name}_q{self.width}_r{trainresult.i}_{step}_eval_test'
            metrics = framework.evaluate(model,
                                         testset,
                                         batch_size=self.batch_size,
                                         dataaugmentations=trainresult.dataaugmentations,
                                         experimenttracking=trainresult.experimenttracking,
                                         dataset_type='test',
                                         name=name)
        return metrics, name

    def __evaluate_and_log(self,  # noqa: PLR0913
                           framework: LearningFramework[T],
                           model: T,
                           trainset: RawData,
                           testset: RawData,
                           trainresult: TrainResult,
                           step: str,
                           step_msg: str) -> dict[str, int | float | numpy.typing.NDArray[Any]]:
        logger.info('Evaluation %s', step_msg)
        metrics, name = self.__evaluate(framework,
                                        model,
                                        trainset,
                                        testset,
                                        trainresult=trainresult,
                                        step=step)
        self.log(QuantizationAwareTrainingLoggerFields(i=trainresult.i,
                                                       name=name,
                                                       step=step,
                                                       width=self.width,
                                                       metrics=metrics))

        return metrics

    def _build_quantized_model(self,
                               model: nn.Module,
                               framework: LearningFramework[Any],
                               model_conf: ModelConfigDict) -> tuple[nn.Module, ModelParamsConfigDict]:
        logger.info('Building quantized model')
        # Create quantized model with parameters from original model
        # Copy params so that original config is not lost for subsequent calls with different models

        qmodel_params: ModelParamsConfigDict = copy.deepcopy(self.qmodel_params)

        if 'update' in model_conf :
            qmodel_params['quant_params'].update(model_conf['update']['quant_params'])

        if 'params' in model_conf:
            qmodel_params = merge_dict(qmodel_params, model_conf['params'])

        self.lsq: bool = qmodel_params['quant_params'].get('LSQ', False)
        self.quant_framework: str | None = qmodel_params['quant_params'].get('quant_framework', None)
        self.width: int = qmodel_params['quant_params'].get('bits', 0)
        self.force_q: int | None = qmodel_params['quant_params'].get('force_q', None)

        prefix = 'Quantized'
        if self.quant_framework is not None :
            prefix = prefix + self.quant_framework
            if self.quant_framework == 'brevitas' :
                import torch
                torch.use_deterministic_algorithms(mode=True)

        # Use model name from config prefixed with Quantized
        quantized_model: nn.Module = getattr(framework.learningmodels, prefix + model_conf['kind'])(**qmodel_params)
        _ = quantized_model.load_state_dict(model.state_dict(), strict=True)
        framework.summary(quantized_model)

        return quantized_model, qmodel_params

    def _update_activation_ranges(self,
                                  framework: PyTorch,
                                  trainresult: TrainResult,
                                  quantized_model: nn.Module) -> nn.Module:
        logger.info('Updating activation ranges')
        name = f'{trainresult.name}_q{self.width}_r{trainresult.i}_update_activation_ranges'
        # Must train to update activation range
        return framework.train(quantized_model,
                        trainset=trainresult.trainset,
                        validationset=trainresult.datamodel.sets.valid,
                        epochs=1,
                        batch_size=self.batch_size,
                        optimizer=None, # Disable optimizer
                        dataaugmentations=trainresult.dataaugmentations,
                        experimenttracking=trainresult.experimenttracking,
                        name=name,
                        precision=32) # Force precision=32 since 'mixed-16' won't work without optimizer, related: https://github.com/Lightning-AI/lightning/issues/17407

    def _quantization_aware_training(self,
                                     framework: LearningFramework[T],
                                     trainresult: TrainResult,
                                     quantized_model: T) -> T:
            logger.info('Performing quantization-aware training for %s epochs', self.epochs)
            # restore optimizer with overwritten params from QAT config
            optimizer = trainresult.optimizer if self.qoptimizer is None else merge_dict(self.qoptimizer, trainresult.optimizer)
            name = f'{trainresult.name}_q{self.width}_r{trainresult.i}_post_train'
            return framework.train(quantized_model,
                            trainset=trainresult.datamodel.sets.train,
                            validationset=trainresult.datamodel.sets.valid,
                            epochs=self.epochs,
                            batch_size=self.batch_size,
                            optimizer=optimizer,
                            dataaugmentations=trainresult.dataaugmentations,
                            experimenttracking=trainresult.experimenttracking,
                            name=name)

    def __export_activation_ranges(self,
                                   model: nn.Module,
                                   save_name: str) -> None:
        from torch.fx.graph import _Namespace, _snake_case

        from qualia_core.learningmodel.pytorch.layers.quantized_layers import quantized_layers
        from qualia_core.learningmodel.pytorch.layers.QuantizedLayer import QuantizedLayer
        from qualia_core.learningmodel.pytorch.Quantizer import Quantizer

        namespace = _Namespace()

        path = Path('out')/'learningmodel'/f'{save_name}_activations_range.txt'
        logger.info('Exporting activation ranges to %s', path)

        with path.open('w') as f:
            for name, m in cast(Generator[tuple[str, nn.Module], None, None], model.named_modules()):
                if len(name) > 0: # _Namespace.create_name tries to access first character without any check for empty string
                    name = _snake_case(name) # convert module name to identifier, '.' replaced with '_', this is what torch.fx uses
                    name = namespace.create_name(name, m)
                if isinstance(m, (QuantizedLayer, *quantized_layers)):
                    q = f'{name},{m.input_q},{m.activation_q},{m.weights_q}'
                    q += f',{m.input_round_mode},{m.activation_round_mode},{m.weights_round_mode}'
                    print(q, file=f)
                    logger.info('%s', q)
                elif not isinstance(m, Quantizer): # Skip Quantizer modules as they need no quantization themselves
                    logger.info("Layer '%s' of type '%s' not quantized", name, type(m))

    @override
    def __call__(self, trainresult: TrainResult, model_conf: ModelConfigDict) -> tuple[ TrainResult, ModelConfigDict]:
        framework = trainresult.framework
        if not isinstance(framework, PyTorch):
            logger.error('Framework %s is not compatible', type(framework))
            raise TypeError

        et = trainresult.experimenttracking

        if et is not None:
            et.start(name=f'{trainresult.name}_q{self.width}_r{trainresult.i}')
            et.hyperparameters = {'model_conf': model_conf,
                                  'qmodel': self.qmodel,
                                  'qoptimizer': self.qoptimizer,
                                  'i': trainresult.i}

        if self._evaluate_before:
            _ = self.__evaluate_and_log(framework=framework,
                                    model=trainresult.model,
                                    trainset=trainresult.trainset,
                                    testset=trainresult.testset,
                                    trainresult=trainresult,
                                    step='baseline',
                                    step_msg='before quantization')

        quantized_model, qmodel_params = self._build_quantized_model(model=trainresult.model,
                                                      framework=framework,
                                                      model_conf=model_conf)

        quantized_model = self._update_activation_ranges(framework=framework,
                                       trainresult=trainresult,
                                       quantized_model=quantized_model)

        metrics = self.__evaluate_and_log(framework=framework,
                                model=quantized_model,
                                trainset=trainresult.trainset,
                                testset=trainresult.testset,
                                trainresult=trainresult,
                                step='PTQ',
                                step_msg='after quantization without training (PTQ)')

        if self.epochs > 0: # Do not attempt quantization-aware training if epochs==0
            quantized_model = self._quantization_aware_training(framework=framework,
                                              trainresult=trainresult,
                                              quantized_model=quantized_model)
            metrics = self.__evaluate_and_log(framework=framework,
                                    model=quantized_model,
                                    trainset=trainresult.trainset,
                                    testset=trainresult.testset,
                                    trainresult=trainresult,
                                    step='QAT',
                                    step_msg='after quantization-aware training')

        model_name = self.process_name(trainresult.name)

        self.__export_activation_ranges(quantized_model, save_name=f'{model_name}_r{trainresult.i}')

        if et is not None:
            et.stop()

        new_model_conf = copy.deepcopy(model_conf)
        new_model_conf['params'] = qmodel_params

        return dataclasses.replace(trainresult, model=quantized_model,
                                    name=model_name,
                                    params=framework.n_params(quantized_model),
                                    mem_params=(framework.n_params(quantized_model) * self.width) // 8,
                                    acc=metrics['testacc'],
                                    metrics=metrics), new_model_conf

    @override
    def process_name(self, name: str) -> str:
        return f'{name}_q{self.width}_force_q{self.force_q if self.force_q is not None else "off"}_e{self.epochs}_LSQ{self.lsq}'

    @override
    def process_mem_params(self, mem_params: int) -> Callable[[LearningFramework[nn.Module | keras.Model],
                                                               nn.Module | keras.Model], int]:
        return lambda framework, model: (framework.n_params(model) * self.width) // 8



from .CNN import CNN
from .MLP import MLP
from .QuantizedCNN import QuantizedCNN
from .QuantizedMLP import QuantizedMLP
from .QuantizedResNet import QuantizedResNet
from .ResNet import ResNet
from .TorchVisionModel import TorchVisionModel

__all__ = ['CNN', 'MLP', 'QuantizedCNN', 'QuantizedMLP', 'QuantizedResNet', 'ResNet',
           'TorchVisionModel']



from abc import ABC, abstractmethod

import torch

from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation


class DataAugmentationPyTorch(ABC, DataAugmentation):
    @abstractmethod
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        ...



import numpy as np

class WindowedSample:
    def __init__(self, features: 'List'=None):
        self.features = features if features is not None else []

    def get_raw_array(self):
        return self.features




from __future__ import annotations

import sys

import torch

from .DataAugmentationPyTorch import DataAugmentationPyTorch

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class GaussianNoise(DataAugmentationPyTorch):

    def __init__(self,
                 sigma: float = 0.375,
                 evaluate: bool = False,  # noqa: FBT001, FBT002
                 before: bool = False,  # noqa: FBT001, FBT002
                 after: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(evaluate=evaluate, before=before, after=after)
        self.sigma = sigma

        self.noisedist = None

    def apply(self, x: torch.Tensor, y: torch.Tensor, device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        if self.noisedist is None: # Lazy init so that CUDA is only used when module is actually called
            self.noisedist = torch.distributions.normal.Normal(torch.tensor(0.0, device=device),
                                                               torch.tensor(self.sigma, device=device))

        noise = self.noisedist.sample(x.shape[-1:]) # Generate noise for each time sample

        noisy_x = x + noise

        return noisy_x, y

    @override
    def __call__(self, data: tuple[torch.Tensor, torch.Tensor], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        return self.apply(*data, device=device)



from __future__ import annotations

import logging
import sys
import time
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import numpy.typing
from qualia_core.datamodel import RawDataModel
from qualia_core.datamodel.RawDataModel import RawData, RawDataSets

from .RawDataset import RawDataset

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

@dataclass
class CIFAR10File:
    data: numpy.typing.NDArray[np.uint8]
    labels: list[int]
    batch_label: bytes
    filenames: list[bytes]

class CIFAR10(RawDataset):
    def __init__(self, path: str='', dtype: str = 'float32') -> None:
        super().__init__()
        self.__path = Path(path)
        self.__dtype = dtype
        self.sets.remove('valid')

    def __unpickle(self, file: Path) -> CIFAR10File:
        import pickle
        with file.open('rb') as fo:
            raw = pickle.load(fo, encoding='bytes')
            content = {k.decode('cp437'): v for k, v in raw.items()}
            return CIFAR10File(**content)

    def __load_train(self, path: Path) -> RawData:
        start = time.time()

        train_x_list: list[numpy.typing.NDArray[np.uint8]] = []
        train_y_list: list[numpy.typing.NDArray[np.int64]] = []
        for i in range(1, 6):
            d = self.__unpickle(path/f'data_batch_{i}')
            train_x_list.append(d.data)
            train_y_list.append(np.array(d.labels))

        train_x_uint8 = np.concatenate(train_x_list)
        train_x_uint8 = train_x_uint8.reshape((train_x_uint8.shape[0], 3, 32, 32)) # N, C, H, W
        train_x_uint8 = train_x_uint8.transpose((0, 2, 3, 1))
        train_x = train_x_uint8.astype(self.__dtype) # N, H, W, C
        train_y = np.concatenate(train_y_list)

        logger.info('__load_train() Elapsed: %s s', time.time() - start)

        return RawData(train_x, train_y)

    def __load_test(self, path: Path) -> RawData:
        start = time.time()

        d = self.__unpickle(path/'test_batch')

        test_x_uint8 = d.data
        test_x_uint8 = test_x_uint8.reshape((test_x_uint8.shape[0], 3, 32, 32)) # N, C, H, W
        test_x_uint8 = test_x_uint8.transpose((0, 2, 3, 1))
        test_x = test_x_uint8.astype(self.__dtype) # N, H, W, C
        test_y = np.array(d.labels)

        logger.info('__load_test() Elapsed: %s s', time.time() - start)

        return RawData(test_x, test_y)

    @override
    def __call__(self) -> RawDataModel:
        return RawDataModel(sets=RawDataSets(
                                    train=self.__load_train(self.__path),
                                    test=self.__load_test(self.__path),
                                ), name=self.name)



import concurrent.futures
import logging
import os
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

import numpy as np
import numpy.typing
from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation
from qualia_core.datamodel.RawDataModel import RawDataModel
from qualia_core.evaluation import Stats
from qualia_core.learningframework.LearningFramework import LearningFramework
from torch.types import Number

logger = logging.getLogger(__name__)

@dataclass
class Result:
    acc: float
    time: float

class Qualia:
    """Custom evaluation loop for Qualia host implementations like Qualia-CodeGen Linux example."""

    def __init__(self) -> None:
        super().__init__()
        self.__deploydir = Path('out')/'deploy'/'Linux'

    def _float_to_hex(self, arr: numpy.typing.NDArray[Any]) -> numpy.typing.NDArray[np.str_]:
        def float_to_hex(x: Number) -> str:
            return float(x).hex()
        return np.vectorize(float_to_hex)(arr)

    def _run_on_split(self,
                       csvdir: Path,
                       tag: str,
                       i: int,
                       test_x: numpy.typing.NDArray[Any],
                       test_y: numpy.typing.NDArray[Any]) -> Result:
        np.savetxt(csvdir/f'testX_{i}.csv', self._float_to_hex(test_x), delimiter=',', fmt='%s')
        np.savetxt(csvdir/f'testY_{i}.csv', test_y, delimiter=',', fmt='%d')

        cmd = [str(self.__deploydir/f'{tag}_Linux'), str(csvdir/f'testX_{i}.csv'), str(csvdir/f'testY_{i}.csv')]
        logger.info('%d Running: %s', i, ' '.join(cmd))

        tstart = time.time() # Start timer
        res = subprocess.run(cmd, # noqa: S603 command line is safe
                             capture_output=True,
                             text=True)
        tstop = time.time() # Stop timer

        if res.stdout:
            logger.info('%d stdout: %s', i, res.stdout)
        if res.stderr:
            logger.info('%d stderr: %s', i, res.stderr)

        accuracy = float(res.stderr.split(' ')[2])

        return Result(accuracy, (tstop - tstart))

    def evaluate(self,
                 framework: LearningFramework,
                 model_kind: str,
                 dataset: RawDataModel,
                 target: str,
                 tag: str,
                 limit: Optional[int] = None,
                 dataaugmentations: Optional[list[DataAugmentation]] = None) -> Stats:
        test_x = dataset.sets.test.x
        test_y = dataset.sets.test.y

        # Apply evaluation "dataaugmentations" to dataset
        if dataaugmentations is not None:
            for da in dataaugmentations:
                if da.evaluate:
                    test_x, test_y = framework.apply_dataaugmentation(da, test_x, test_y)

        # create log directory
        (Path('logs')/'evaluate'/target).mkdir(parents=True, exist_ok=True)

        # create data CSV dir
        csvdir = Path('out')/'data'/dataset.name/'csv'
        csvdir.mkdir(parents=True, exist_ok=True)

        # Flatten test vectors
        test_x = test_x.reshape((test_x.shape[0], -1))

        if limit:
            test_x = test_x[:limit]
            test_y = test_y[:limit]

        test_vector_count = test_y.shape[0]

        # Split test into chunks
        cpu = os.cpu_count()
        chunks = cpu if cpu is not None else 2
        test_x = np.array_split(test_x, chunks)
        test_y = np.array_split(test_y, chunks)

        with concurrent.futures.ProcessPoolExecutor(max_workers=chunks) as executor:
            futures = [executor.submit(self._run_on_split, csvdir, tag, i, x, y) for i, (x, y) in enumerate(zip(test_x, test_y))]
            results = [f.result() for f in futures]

        avg_time = sum(r.time for r in results) / test_vector_count
        accuracy = sum(r.acc * len(y) for r, y in zip(results, test_y)) / test_vector_count

        return Stats(avg_time=avg_time, accuracy=accuracy)

        # avg it/secs
        # ram usage
        # rom usage
        # cpu type
        # cpu model
        # cpu freq
        # accuracy
        # power consumption



from tensorflow.keras.optimizers import Optimizer
import tensorflow as tf

class DummyOptimizer(Optimizer):
    def __init__(self, name='DSOMOptimizer', **kwargs):
        super().__init__(name, **kwargs)

    def _create_slots(self, var_list):
        pass

    def minimize(self):
        pass

    def apply_gradients(self, grads_and_vars, name=None):
        pass

    def get_config(self):
        return super().get_config()



import numpy as np

from .Preprocessing import Preprocessing

class BandPassFilter(Preprocessing):
    '''Band pass filter
       Requires fixed sampling frequency, do an interpolation before.
    '''
    def __init__(self, f1: float, f2: float, fs: float, sensorkind, dimension: str):
        self.__f1 = f1
        self.__f2 = f2
        self.__fs = fs
        self.__dimension = dimension

        if isinstance(sensorkind, str):
            import qualia_core.datamodel.sensor as sensors
            sensorkind = getattr(sensors, sensorkind)
        self.__sensorkind = sensorkind

    def __filter(self, array):
        import scipy.signal

        nyq = 0.5 * self.__fs
        w1 = self.__f1 / nyq
        w2 = self.__f2 / nyq
        b, a = scipy.signal.butter(1, [w1, w2], btype='bandpass')
        if len(array) < len(a) * len(b): # Do not apply filter if data smaller than filter
            print(f'Warning: band pass filter not applied for current activity. Signal len {len(array)} smaller than filter length {len(a) * len(b)}')
            return array
        else:
            return scipy.signal.filtfilt(b, a, array)

    def __call__(self, datamodel):

        for name, subjectset in datamodel:
            for subject in subjectset:
                for activity in subject.activities:
                    signal = np.array([getattr(sensor, self.__dimension) for sample in activity.samples
                                                for sensor in sample.sensors if isinstance(sensor, self.__sensorkind)])
                    filtered = self.__filter(signal)
                    ifiltered = np.nditer(filtered, order='A') # using default 'K' order will reverse element, probably a side-effect
                                                               # of filtfilt

                    for sample in activity.samples:
                        for sensor in sample.sensors:
                            if not isinstance(sensor, self.__sensorkind):
                                continue
                            setattr(sensor, self.__dimension, next(ifiltered))
        return datamodel



from pathlib import Path
import re
import sys
from qualia_core.utils.process import subprocesstee                                                                                     
from qualia_core.evaluation.Stats import Stats

class STM32CubeAI:
    def __init__(self,
        stm32cubeai_args=tuple(),
        mode=''):
        self.__mode = mode
        self.__stm32cubeai_args = stm32cubeai_args

        # Project not compatible with 6.0.0 yet, and 6.0.0 is buggy anyway. Use 5.2.0
        self.__stm32cubeai_bin = Path.home()/'STM32Cube'/'Repository'/'Packs'/'STMicroelectronics'/'X-CUBE-AI'/'5.2.0'/'Utilities'/'linux'/'stm32ai'
        #self.__stm32cubeai_bin = next((Path.home()/'STM32Cube'/'Repository'/'Packs'/'STMicroelectronics'/'X-CUBE-AI').glob('*'))/'Utilities'/'linux'/'stm32ai'

    def __dataset_to_csv(self, dataset, outdir, limit):
        import numpy as np
        testX = dataset.sets.test.x
        testY = dataset.sets.test.y
        if limit:
            testX = testX[:limit]
            testY = testY[:limit]
        testX = testX.reshape((testX.shape[0], -1))
        testY = testY.reshape((testY.shape[0], -1))
        np.savetxt(outdir/'testX.csv', testX, delimiter=",", fmt='%f')
        np.savetxt(outdir/'testY.csv', testY, delimiter=",", fmt='%f')

    def __validate(self, testX: Path, testY: Path, modelpath: Path, compression: int, outdir: Path, logdir: Path, tag: str):

        cmd = str(self.__stm32cubeai_bin)
        args =  ('validate',
                    '--name', 'network',
                    '--model', str(modelpath),
                    '--mode', self.__mode,
                    #'--type', 'keras', # automatically detected, using tflite for now
                    #'--compression', str(compression),
                    '--valinput', str(testX),
                    '--valoutput', str(testY),
                    '--verbosity', '1',
                    '--workspace', str(outdir/'workspace'),
                    '--output', str(outdir/tag),
                    '--classifier'
                ) + self.__stm32cubeai_args
        print(cmd, *args)
        with (logdir/f'{tag}.txt').open('wb') as logfile:
            logfile.write(' '.join([str(cmd), *args, '\n']).encode('utf-8'))
            returncode, outputs = subprocesstee.run(cmd, *args, files={sys.stdout: logfile, sys.stderr: logfile})
        return returncode, outputs

    def __parse_validate_stdout(self, s: str):
        duration = re.search("^\s*duration\s*:\s*([.\d]+)\s+ms\s+\(average\)$", s, re.MULTILINE)
        if duration is not None:
            duration = float(duration.group(1))/1000

        accuracy = re.search(f'^{self.__mode}\ C-model\ #1\s+([.\d]+)%.*$', s, re.MULTILINE)
        if accuracy is not None:
            accuracy = float(accuracy.group(1))/100

        return duration, accuracy

    def evaluate(self, framework, model_kind, dataset, target: str, tag: str, limit: int=None, dataaugmentations=[]):
        if dataaugmentations:
            raise ValueError(f'dataaugmentations not supported for {self.__class__.__name__}')

        outdir = Path('out')/'deploy'/target
        logdir = Path('out')/'evaluate'/target
        logdir.mkdir(parents=True, exist_ok=True)

        self.__dataset_to_csv(dataset, logdir, limit)

        return_code, outputs = self.__validate(testX=logdir/'testX.csv',
                                                testY=logdir/'testY.csv',
                                                modelpath=Path('out')/'deploy'/'stm32cubeai'/f'{tag}.tflite',
                                                compression=None,
                                                outdir=outdir,
                                                logdir=logdir,
                                                tag=tag)
        if return_code != 0:
            return None
        duration, accuracy = self.__parse_validate_stdout(outputs[1].decode())

        return Stats(avg_time=duration, accuracy=accuracy)



from dataclasses import dataclass, field

import numpy as np
import numpy.typing

from qualia_core.datamodel.TimeSample import TimeSample

from .Activities import Activities


@dataclass
class Activity:
    kind: Activities
    samples: list[TimeSample] = field(default_factory=list)

    def get_raw_array(self) -> list[list[numpy.typing.NDArray[np.float32]]]:
        return [s.get_raw_array() for s in self.samples]



from __future__ import annotations

from collections import namedtuple
from importlib.resources import files
from pathlib import Path

from qualia_core.deployment.Deployer import Deployer
from qualia_core.utils.process import subprocesstee


class SparkFunEdge(Deployer):
    import qualia_core.evaluation.target.Qualia as evaluator  # Suggested evaluator

    def __init__(self,
                 dev: str = '/dev/ttyUSB0',
                 modeldir: Path = Path('out')/'qualia_codegen',
                 projectdir: Path | None = None,
                 outdir: Path = Path('out')/'deploy'/'SparkFunEdge') -> None:
        super().__init__()
        self.__dev = dev
        self.__modeldir = modeldir
        self.__projectdir = projectdir if projectdir is not None else files('qualia_codegen_core.examples')/'SparkFunEdge'
        self.__outdir = outdir

    def __run(self, cmd, *args):
        print(cmd, *args)
        returncode, outputs = subprocesstee.run(str(cmd), *args)
        return returncode == 0

    def __create_outdir(self):
        self.__outdir.mkdir(parents=True, exist_ok=True)

    def __clean(self, tag: str, model):
        modeldir = self.__modeldir/model.name
        outdir = self.__outdir/tag
        return self.__run('make',
                          '-C', str(self.__projectdir),
                          f'MODELDIR={modeldir.absolute()}',
                          f'OUT={outdir.absolute()}',
                          'clean'
                        )

    def __build(self, tag: str, model, optimize: str):
        modeldir = self.__modeldir/model.name
        outdir = self.__outdir/tag
        args = ['-C', str(self.__projectdir),
                f'MODELDIR={modeldir.absolute()}',
                f'OUT={outdir.absolute()}']
        if optimize == 'cmsis-nn':
            args.append('WITH_CMSIS_NN=1')
        return self.__run('make', *args)

    def __upload(self, tag: str):
        outdir = self.__outdir/tag
        return self.__run('make',
                          '-C', str(self.__projectdir),
                          f'OUT={outdir.absolute()}',
                          f'SERIAL_PORT={self.__dev}',
                          'bootload'
                        )

    def prepare(self, tag, model, optimize: str, compression: int):
        if optimize and optimize != 'cmsis-nn':
            raise ValueError(f'Optimization {optimize} not available for {self.__class__.__name__}')
        if compression != 1:
            raise ValueError(f'No compression available for {self.__class__.__name__}')

        print('model:', model)

        self.__create_outdir()

        self.__clean(tag=tag, model=model)
        if not self.__build(tag=tag, model=model, optimize=optimize):
            return None
        return self

    def deploy(self, tag):
        input('Put target in programming mode and press Enter…')
        self.__upload(tag=tag)

        return namedtuple('Deploy', ['rom_size', 'ram_size', 'evaluator'])(self._rom_size(tag), self._ram_size(tag), self.evaluator)

    def _rom_size(self, tag):
        return super()._rom_size(self.__outdir/tag/'main.axf', 'arm-none-eabi-size')

    def _ram_size(self, tag):
        return super()._ram_size(self.__outdir/tag/'main.axf', 'arm-none-eabi-size')



from string import Template
from pathlib import Path

class TFLite2CArray:
    def __init__(self):
        with (Path(__file__).parent.parent.parent/'assets'/'template'/'tflite_model_data.cc').open('r') as f:
            self.__templatecc = Template(f.read())

    def convert(self, data, input_shape):
        datax = [hex(b) for b in data]
        self.__output = ', '.join(datax)
        self.__datalen = len(datax)
        self.__dims = ', '.join((str(v) if v != None else '1' for v in input_shape)) # convert Tuple to string replacing None by 1
        self.__dimslen = len(input_shape)
        return self

    @property
    def cc(self):
        return self.__templatecc.substitute(data=self.__output, datalen=self.__datalen, dims=self.__dims, dimslen=self.__dimslen)



from enum import Enum

# From https://github.com/irgeek/StrEnum
class StrEnum(str, Enum):
    def __str__(self):
        return self.value

    def _generate_next_value_(name, *_):
        return name



from .test_qualia_codegen_linux import TestQualiaCodeGenLinux
from .test_qualia_codegen_nucleol452rep import TestQualiaCodeGenNucleoL452REP
from .test_qualia_codegen_sparkfunedge import TestQualiaCodeGenSparkFunEdge
from .test_stm32cubeai_nucleol452rep import TestSTM32CubeAINucleoL452REP
from .test_tflitemicro_sparkfunedge import TestTFLiteMicroSparkFunEdge

__all__ = ['TestQualiaCodeGenLinux', 'TestQualiaCodeGenNucleoL452REP', 'TestQualiaCodeGenSparkFunEdge', 'TestSTM32CubeAINucleoL452REP',
           'TestTFLiteMicroSparkFunEdge']



from __future__ import annotations

import importlib
import importlib.util
import logging
import sys
from pathlib import Path
from typing import Any, Callable, NoReturn

import numpy as np
import numpy.typing
import torch
import torch.distributed
import torch.utils.data
from torch import nn

from qualia_core.experimenttracking.pytorch.ExperimentTrackingPyTorch import ExperimentTrackingPyTorch
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger import TextLogger

from .LearningFramework import LearningFramework

if TYPE_CHECKING:
    from pytorch_lightning.loggers import Logger  # noqa: TCH002
    from pytorch_lightning.trainer.connectors.accelerator_connector import _PRECISION_INPUT  # noqa: TCH002

    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.dataaugmentation.pytorch.DataAugmentationPyTorch import DataAugmentationPyTorch  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawData  # noqa: TCH001
    from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking  # noqa: TCH001
    from qualia_core.typing import OptimizerConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class PyTorch(LearningFramework[nn.Module]):
    # Reference framework-specific external modules
    from pytorch_lightning import LightningModule

    import qualia_core.dataaugmentation.pytorch
    import qualia_core.experimenttracking.pytorch
    import qualia_core.learningmodel.pytorch

    dataaugmentations = qualia_core.dataaugmentation.pytorch
    experimenttrackings = qualia_core.experimenttracking.pytorch
    learningmodels = qualia_core.learningmodel.pytorch

    trainer = None

    class TrainerModule(LightningModule):
        def __init__(self,  # noqa: PLR0913
                     model: nn.Module,
                     max_epochs: int = 0,
                     optimizer: OptimizerConfigDict | None = None,
                     dataaugmentations: list[DataAugmentationPyTorch] | None = None,
                     num_classes: int | None = None,
                     experimenttracking_init: Callable[[], NoReturn] | None = None) -> None:
            super().__init__()
            self.model = model
            self.max_epochs = max_epochs
            self.optimizer = optimizer
            self.dataaugmentations = dataaugmentations if dataaugmentations is not None else []
            self.experimenttracking_init = experimenttracking_init

            self.configure_loss()
            self.configure_metrics(num_classes=num_classes)

        @override
        def setup(self, stage: str) -> None:
            super().setup(stage)

            if stage == 'fit':
                if '16' in self.trainer.precision:
                    torch.set_float32_matmul_precision('medium')
                else:
                    torch.set_float32_matmul_precision('high')

            # Required in some cases with ddp_spawn to connect current process with experimenttracking task
            if self.experimenttracking_init is not None:
                self.experimenttracking_init()

        def configure_loss(self) -> None:
            from qualia_core.dataaugmentation.pytorch import Mixup

            self.crossentropyloss = nn.CrossEntropyLoss()
            self.softmax = nn.Softmax(dim=1)

            self.mixup = next((da for da in self.dataaugmentations if isinstance(da, Mixup)), None) # Check if Mixup is enabled
            if self.mixup:
                self.loss = self.mixup.loss.__get__(self.mixup)
            else:
                self.loss = self.crossentropyloss

        def configure_metrics(self, num_classes: int | None = None) -> None:
            import torchmetrics

            metrics = torchmetrics.MetricCollection({
                'prec': torchmetrics.Precision(task='multiclass', average='macro', num_classes=num_classes),
                'rec': torchmetrics.Recall(task='multiclass', average='macro', num_classes=num_classes),
                'f1': torchmetrics.F1Score(task='multiclass', average='macro', num_classes=num_classes),
                'acc': torchmetrics.Accuracy(task='multiclass', average='micro', num_classes=num_classes),
                'avgclsacc': torchmetrics.Accuracy(task='multiclass', average='macro', num_classes=num_classes)})

            self.train_metrics = metrics.clone(prefix='train')
            self.val_metrics = metrics.clone(prefix='val')
            self.test_metrics = metrics.clone(prefix='test')

        @override
        def on_before_batch_transfer(self,
                                    batch: tuple[torch.Tensor, torch.Tensor],
                                    dataloader_idx: int) -> tuple[torch.Tensor, torch.Tensor]:
            if self.dataaugmentations:
                for da in self.dataaugmentations:
                    if (self.trainer.training or da.evaluate) and da.before:
                        batch = da(batch, device=self.device)
            return batch

        @override
        def on_after_batch_transfer(self,
                                    batch: tuple[torch.Tensor, torch.Tensor],
                                    dataloader_idx: int) -> tuple[torch.Tensor, torch.Tensor]:
            if self.dataaugmentations:
                for da in self.dataaugmentations:
                    if (self.trainer.training or da.evaluate) and da.after:
                        batch = da(batch, device=self.device)
            return batch

        @override
        def training_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_nb: int) -> torch.Tensor:
            x, y = batch
            logits = self(x)

            if not self.mixup:
                self.train_metrics(self.softmax(logits), y)
                self.log_dict(self.train_metrics)

            loss = self.loss(logits, y)
            self.log('train_loss',loss)
            return loss

        @override
        def validation_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_nb: int) -> None:
            x, y = batch
            logits = self.softmax(self(x)) # lightning 1.2 requires preds in [0, 1]

            self.val_metrics(logits, y)
            self.log_dict(self.val_metrics, prog_bar=True)

        @override
        def test_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_nb: int) -> None:
            x, y = batch
            logits = self.softmax(self(x)) # lightning 1.2 requires preds in [0, 1]

            self.test_metrics(logits, y)
            self.log_dict(self.test_metrics, prog_bar=True)

        @override
        def predict_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int, dataloader_idx: int = 0) -> torch.Tensor:
            x, y = batch
            preds = self(x)

            ### Below is boilerplate code to support distributed inference if needed ###
            # By default this will not be used since we set devices=1 in Trainer of PyTorch LearningFramework test()
            shape = torch.tensor(preds.shape, device=preds.device)
            # Fetch shapes of predictions across all nodes
            all_shapes: torch.Tensor = self.all_gather(shape)
            # Apparently only one shape gathered, most likely running on single node so no preds to gather
            if torch.equal(shape, all_shapes):
                return preds

            # Resize preds to fit largest shape since preds will be gathered in a single tensor in a new dim
            # Will be resized back after gathering
            max_shape: torch.Tensor = torch.max(all_shapes, dim=0)[0]
            preds.resize_(*max_shape)

            # Fetch predictions across all nodes
            all_preds: torch.Tensor = self.all_gather(preds)
            # Split to list among the new dimension (1 row per node)
            preds_list = torch.split(all_preds, 1)
            # Resize each preds tensor to its original size
            for shape, t in zip(all_shapes, preds_list):
                t.resize_(*shape)
            # Concat in the same dimension
            return torch.cat(preds_list)

        @override
        def forward(self, x: torch.Tensor) -> torch.Tensor:
            return self.model(x)

        @override
        def on_train_epoch_end(self) -> None:
            super().on_train_epoch_end()

            if not self.mixup:
                self.log_dict(self.train_metrics, prog_bar=True, sync_dist=True)
            for param_group in self.trainer.optimizers[0].param_groups:
                self.log('lr', param_group['lr'], prog_bar=True, sync_dist=True)

            # New line after each train epoch if no validation step
            if not self.trainer.val_dataloaders:
                print()

        @override
        def on_validation_epoch_end(self) -> None:
            super().on_validation_epoch_end()
            # New line after validation step at the end of each epoch
            print()

        @override
        def configure_optimizers(self) -> tuple[list[torch.optim.Optimizer] | list[torch.optim.lr_scheduler.LRScheduler]]:
            import torch.optim as optimizers
            if importlib.util.find_spec('torch_optimizer') is None:
                print('Warning: torch_optimizer not found, not importing additional optimizers', file=sys.stderr)
            else:
                optimizers.__dict__.update(importlib.import_module('torch_optimizer').__dict__) # Merge additional torch_optimizer
            import torch.optim.lr_scheduler as schedulers

            if not self.optimizer:
                return None

            optimizer = getattr(optimizers, self.optimizer['kind'])(self.parameters(), **self.optimizer.get('params', {}))
            print('Optimizer:', optimizer, self.optimizer.get('params', {}))
            if 'scheduler' in self.optimizer:
                try:
                    scheduler = getattr(schedulers, self.optimizer['scheduler']['kind'])(optimizer, **self.optimizer['scheduler']['params'])
                except AttributeError:
                    import qualia_core.learningframework.CustomScheduler as customschedulers
                    scheduler = getattr(customschedulers, self.optimizer['scheduler']['kind'])(optimizer, **self.optimizer['scheduler']['params'])

                print('Scheduler:', scheduler, self.optimizer['scheduler']['params'])
                return [optimizer], [scheduler]
            return [optimizer], []

    def __init__(self,
                 use_best_epoch: bool = False,
                 enable_progress_bar: bool = True,
                 progress_bar_refresh_rate: int = 1,
                 accelerator: str = 'auto',
                 devices: int | str | list[int] = 'auto',
                 precision: _PRECISION_INPUT = 32) -> None:
        super().__init__()
        self._use_best_epoch = use_best_epoch
        self._enable_progress_bar = enable_progress_bar
        self._progress_bar_refresh_rate = progress_bar_refresh_rate
        self.accelerator = accelerator
        self.devices = devices
        self.precision = precision

        self.log = TextLogger(name=__name__)

        # Force using 'spawn' instead of 'popen' start_method for ddp strategy in case of multiple devices.
        # This is done to avoid starting our 'qualia' script all over again which causes some issues
        # since it does much more than just perform the training, this would result in all the steps being performed again
        # including duplicating tests, postprocessing, etc…
        # This is probably significantly slower than using 'popen' (and may even lead to longer run time than single device)
        # but at least it allows running multi-GPU training by default, and running single device exclusively can still be done by
        # settings learningframework.devices to 1 or exposing only one device with CUDA_VISIBLE_DEVICES.
        # We modify the registry so that we can still use strategy='auto' in the trained to use the SingleDevice strategy in case
        # only one device is being used since this is much faster than initializing DDP.
        from pytorch_lightning.strategies import StrategyRegistry
        from pytorch_lightning.strategies.ddp import DDPStrategy
        start_method = 'spawn'
        StrategyRegistry.register('ddp',
                                   DDPStrategy,
                                   description=f'DDP strategy with `start_method={start_method!r}`',
                                   start_method='spawn',
                                   override=True)

    @staticmethod
    def channels_last_to_channels_first(x: numpy.typing.NDArray[Any]) -> numpy.typing.NDArray[Any]:
        if len(x.shape) == 4:
            x = x.transpose(0, 3, 1, 2)
        elif len(x.shape) == 3:
            x = x.swapaxes(1, 2)
        else:
            raise ValueError(f'Unsupported number of axes in dataset: {len(x.shape)}, must be 3 or 4')
        return x

    @staticmethod
    def channels_first_to_channels_last(x: numpy.typing.NDArray[Any]) -> numpy.typing.NDArray[Any]:
        if len(x.shape) == 4:
            x = x.transpose(0, 2, 3, 1)
        elif len(x.shape) == 3:
            x = x.swapaxes(2, 1)
        else:
            raise ValueError(f'Unsupported number of axes in dataset: {len(x.shape)}, must be 3 or 4')
        return x


    class DatasetFromArray(torch.utils.data.Dataset[tuple[numpy.typing.NDArray[np.float32], numpy.typing.NDArray[np.int32]]]):
        def __init__(self, dataset: RawData) -> None:
            super().__init__()
            self.x = PyTorch.channels_last_to_channels_first(dataset.x)
            self.y = dataset.y.argmax(axis=-1)
            self.y = np.where(np.isnan(dataset.y).any(axis=-1),
                -1,
                self.y)

        def __len__(self) -> int:
            return len(self.x)

        @override
        def __getitem__(self, index: int) -> tuple[numpy.typing.NDArray[np.float32], numpy.typing.NDArray[np.int32]]:
            return self.x[index], self.y[index]

    def logger(self,
               experimenttracking: ExperimentTracking | None,
               name: str) -> list[Logger]:
        from pytorch_lightning.loggers import CSVLogger, TensorBoardLogger
        loggers: list[Logger] = [CSVLogger(save_dir='logs/PyTorchLightning', name=name)]
        if importlib.util.find_spec('tensorboard') is None and importlib.util.find_spec('tensorboardX') is None:
            logger.warning('tensorboard or tensorboardX not found, disabling TensorBoardLogger')
        else:
            loggers.append(TensorBoardLogger(save_dir='lightning_logs', name=name))
        if isinstance(experimenttracking, ExperimentTrackingPyTorch) and experimenttracking.logger is not None:
            loggers.append(experimenttracking.logger)
        return loggers

    @override
    def train(self,  # noqa: PLR0913
              model: nn.Module,
              trainset: RawData | None,
              validationset: RawData | None,
              epochs: int,
              batch_size: int,
              optimizer: OptimizerConfigDict | None,
              dataaugmentations: list[DataAugmentation] | None = None,
              experimenttracking: ExperimentTracking | None = None,
              name: str | None = None,
              precision: _PRECISION_INPUT | None = None) -> nn.Module:
        import os

        from pytorch_lightning import Trainer, seed_everything
        from pytorch_lightning.callbacks import ModelCheckpoint, TQDMProgressBar
        from torch.utils.data import DataLoader

        # PyTorch-Lightning >= 1.3.0 resets seed before training, increment seed between trainings to get different values between experiments
        seed = os.environ.get("PL_GLOBAL_SEED", None)
        if seed is None:
            print("Warning: PyTorch not seeded", file=sys.stderr)
        else:
            seed_everything(int(seed) * 100)

        checkpoint_callback = ModelCheckpoint(dirpath=f"out/checkpoints/{name}", save_top_k=2, monitor="valavgclsacc", mode="max")
        callbacks = [checkpoint_callback]
        if self._enable_progress_bar:
            callbacks.append(TQDMProgressBar(refresh_rate=self._progress_bar_refresh_rate))

        experimenttracking_init = experimenttracking.initializer if experimenttracking is not None else None

        trainer = Trainer(max_epochs=epochs,
                          accelerator=self.accelerator,
                          devices=self.devices,
                          precision=self.precision if precision is None else precision,
                          deterministic=True,
                          logger=self.logger(experimenttracking, name=name),
                          enable_progress_bar=self._enable_progress_bar,
                          callbacks=callbacks)
        trainer_module = self.TrainerModule(model,
                                            max_epochs=epochs,
                                            optimizer=optimizer,
                                            dataaugmentations=dataaugmentations,
                                            num_classes=trainset.y.shape[-1],
                                            experimenttracking_init=experimenttracking_init)
        #self.trainer.fit(trainer_module,
        #                    DataLoader(self.DatasetFromTF(trainset), batch_size=None), [
        #                        DataLoader(self.DatasetFromTF(originalset), batch_size=None),
        #                        DataLoader(self.DatasetFromTF(validationset), batch_size=None)
        #                    ])
        # Bug in PyTorch-Lightning 1.0 with multiple dataloaders

        # Do not attempt to train with no epochs
        if epochs < 1:
            return model

        print('Epochs:', epochs, trainer.max_epochs)
        trainer.fit(trainer_module,
                            DataLoader(self.DatasetFromArray(trainset), batch_size=batch_size,shuffle=True),
                            #num_workers=2, persistent_workers=True, pin_memory=True),
                            DataLoader(self.DatasetFromArray(validationset), batch_size=batch_size) if validationset is not None else None
                        )

        if self._use_best_epoch:
            print(f'Loading back best epoch: {checkpoint_callback.best_model_path}, score: {checkpoint_callback.best_model_score}')
            trainer_module = self.TrainerModule.load_from_checkpoint(checkpoint_callback.best_model_path,
                                                                    model=model,
                                                                    optimizer=optimizer,
                                                                    dataaugmentations=dataaugmentations,
                                                                    num_classes=trainset.y.shape[-1])
            model = trainer_module.model
        return model

    @override
    def evaluate(self,
                 model: nn.Module,
                 testset: RawData,
                 batch_size: int,
                 dataaugmentations: list[DataAugmentation],
                 experimenttracking: ExperimentTracking | None = None,
                 dataset_type: str = '',
                 name: str = '') -> dict[str, int | float | numpy.typing.NDArray[Any]]:
        import torch
        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import TQDMProgressBar
        from torch.utils.data import DataLoader

        self.log(f'{name=}')
        self.log(f'{dataset_type=}')

        # Force testing on single device to avoid issues caused by DDP generating different batches
        # Either first of given devices list or first default
        devices = self.devices[:1] if isinstance(self.devices, list) else 1

        callbacks = []
        if self._enable_progress_bar:
            callbacks.append(TQDMProgressBar(refresh_rate=self._progress_bar_refresh_rate))

        trainer = Trainer(max_epochs=0,
                          accelerator=self.accelerator,
                          devices=devices,
                          deterministic=True,
                          logger=self.logger(experimenttracking, name=name),
                          enable_progress_bar=self._enable_progress_bar,
                          callbacks=callbacks)
        trainer_module = self.TrainerModule(model, dataaugmentations=dataaugmentations, num_classes=testset.y.shape[-1])
        metrics = trainer.test(trainer_module, DataLoader(self.DatasetFromArray(testset), batch_size=batch_size))
        predictions = torch.cat(trainer.predict(trainer_module, DataLoader(self.DatasetFromArray(testset), batch_size=batch_size)))
        # predict returns list of predictions according to batches
        self.log(f'{metrics=}')

        print('Confusion matrix:')
        cm = self.confusion_matrix(predictions, testset, device=trainer_module.device)
        ncm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

        avg_class_accuracy = ncm.diagonal().mean()
        print(f'{avg_class_accuracy=}')
        self.log(f'{avg_class_accuracy=}')

        with np.printoptions(threshold=sys.maxsize, suppress=True, linewidth=sys.maxsize, precision=2):
            print(cm)
            print("Normalized:")
            print(ncm)

            self.log(f'{cm=}')
            self.log(f'{ncm=}')

            if experimenttracking is not None and experimenttracking.logger is not None:
                experimenttracking.logger.experiment['cm'].log(np.array2string(cm))
                experimenttracking.logger.experiment['ncm'].log(np.array2string(ncm))
        metrics[0]['cm'] = cm
        metrics[0]['ncm'] = ncm

        return metrics[0]

    @override
    def predict(self,  # noqa: PLR0913
                 model: nn.Module,
                 dataset: RawData,
                 batch_size: int,
                 dataaugmentations: list[DataAugmentation],
                 experimenttracking: ExperimentTracking | None = None,
                 name: str = '') -> torch.Tensor:
        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import Callback, TQDMProgressBar
        from torch.utils.data import DataLoader

        # Force testing on single device to avoid issues caused by DDP generating different batches
        # Either first of given devices list or first default
        devices = self.devices[:1] if isinstance(self.devices, list) else 1

        callbacks: list[Callback] = []
        if self._enable_progress_bar:
            callbacks.append(TQDMProgressBar(refresh_rate=self._progress_bar_refresh_rate))

        trainer = Trainer(max_epochs=0,
                          accelerator=self.accelerator,
                          devices=devices,
                          deterministic=True,
                          logger=self.logger(experimenttracking, name=name),
                          enable_progress_bar=self._enable_progress_bar,
                          callbacks=callbacks)
        trainer_module = self.TrainerModule(model,
                                            dataaugmentations=dataaugmentations,
                                            num_classes=dataset.y.shape[-1])
        predictions = trainer.predict(trainer_module,
                               DataLoader(self.DatasetFromArray(dataset), batch_size=batch_size))

        if predictions is None:
            logger.error('predict() returned None')
            raise RuntimeError
        if not isinstance(predictions[0], torch.Tensor):
            logger.error('Expected predict() result to be a list of torch.Tensor, got: %s', type(predictions[0]))
            raise TypeError

        return torch.cat(predictions)

    def confusion_matrix(self, predictions: torch.Tensor, testset: RawData, device: str) -> numpy.typing.NDArray[np.int32]:
        import torch
        from torchmetrics import ConfusionMatrix

        testy = torch.tensor(testset.y.argmax(axis=1), device=device)

        confmat = ConfusionMatrix(task='multiclass', num_classes=testset.y.shape[1])
        return confmat(predictions.to(device=device).argmax(axis=1), testy).int().numpy()

    @override
    def load(self, name: str, model: nn.Module) -> nn.Module:
        import torch
        path = Path('out')/'learningmodel'/f'{name}.pth'
        if path.is_file():
            state_dict = torch.load(path)
            model.load_state_dict(state_dict)
            logger.info('Loaded %s.', path)
        else:
            logger.warning('%s not found, not loading weights.', path)
        return model

    @override
    def export(self, model: nn.Module, name: str) -> None:
        import torch
        outdir = Path('out')/'learningmodel'
        outdir.mkdir(parents=True, exist_ok=True)
        torch.save(model.state_dict(), outdir/f'{name}.pth')

    @override
    def summary(self, model: nn.Module) -> None:
        """Print model summary."""
        print(model)
        print(f'Number of parameters: {self.n_params(model)}')

    @override
    def n_params(self, model: nn.Module) -> int:
        import numpy as np
        return np.sum([params.numel() for params in model.parameters()])

    @override
    def save_graph_plot(self, model: nn.Module, model_save: str) -> None:
        import sys
        print('Warning: saving PyTorch graph plot unsupported', file=sys.stderr)

    @override
    def apply_dataaugmentation(self,
                               da: DataAugmentationPyTorch,
                               x: numpy.typing.NDArray[Any],
                               y: numpy.typing.NDArray[Any],
                               device: torch.device | None = None) -> tuple[numpy.typing.NDArray[Any], numpy.typing.NDArray[Any]]:
        import torch
        x = self.channels_last_to_channels_first(x)
        tensor_x = torch.tensor(x, device=device)
        tensor_y = torch.tensor(y, device=device)
        tensor_x, tensor_y = da((tensor_x, tensor_y), device=device)
        x = tensor_x.numpy()
        y = tensor_y.numpy()
        x = self.channels_first_to_channels_last(x)
        return x, y



from __future__ import annotations

from qualia_core.typing import TYPE_CHECKING

from .CSVFormatter import CSVFormatter
from .Logger import Logger, T

if TYPE_CHECKING:
    from pathlib import Path  # noqa: TCH003


class CSVLogger(Logger[T]):
    def __init__(self,
                 name: str,
                 file: Path | None = None,
                 fields: type[T] | None = None,
                 log_fields: bool = True) -> None:  # noqa: FBT001, FBT002
        super().__init__(name=name, file=file, suffix='.csv', fields=fields, log_fields = log_fields, formatter=CSVFormatter())



from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, NamedTuple

import colorful as cf  # type: ignore[import-untyped]

from qualia_core.qualia import train
from qualia_core.typing import TYPE_CHECKING
from qualia_core.utils.logger import CSVLogger, Logger

if TYPE_CHECKING:
    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001
    from qualia_core.typing import ConfigDict
    from qualia_core.utils import Git  # noqa: TCH001
    from qualia_core.utils.plugin import QualiaComponent  # noqa: TCH001

logger = logging.getLogger(__name__)

class LearningModelLoggerFields(NamedTuple):
    name: str
    i: int
    params: int
    mem_params: int
    accuracy: float

class Train:
    def __call__(self,  # noqa: PLR0913
                 qualia: QualiaComponent,
                 learningframework: LearningFramework[Any],
                 dataaugmentations: list[DataAugmentation],
                 data: RawDataModel,
                 config: ConfigDict,
                 git: Git) -> dict[str, Logger[LearningModelLoggerFields]]:
        loggers: dict[str, Logger[LearningModelLoggerFields]] = {}

        log: CSVLogger[LearningModelLoggerFields] = CSVLogger('learningmodel',
                                                              file=Path('learningmodel')
                                                                   /f'{git.short_hash()}_{datetime.now():%Y-%m-%d_%H-%M-%S}.csv')
        loggers['learningmodel'] = log
        log.fields = LearningModelLoggerFields # Write column names

        experimenttracking = config.get('experimenttracking', None)

        (Path('out')/'learningmodel').mkdir(parents=True, exist_ok=True)

        for i in range(config['bench']['first_run'], config['bench']['last_run']+1):
            for m in config['model']:
                if m.get('disabled', False):
                    continue

                if m.get('load', False):
                    print(f'{cf.bold}Loading {cf.blue}{m["name"]}{cf.close_fg_color}, run {cf.red}{i}{cf.reset}')
                else:
                    print(f'{cf.bold}Training {cf.blue}{m["name"]}{cf.close_fg_color}, run {cf.red}{i}{cf.reset}')
                print(f'{cf.bold}Params:{cf.reset} {m=}')

                et = None
                if experimenttracking is not None:
                    et = getattr(learningframework.experimenttrackings, experimenttracking['kind'])(**experimenttracking.get('params', {}))
                    et.start()
                    et.hyperparameters = {'config': config, 'model': m, 'i': i}

                model = getattr(learningframework.learningmodels, m['kind'], None)
                if model is None:
                    logger.error("Could not load model.kind '%s' from learningmodels '%s'.",
                                 m['kind'],
                                 learningframework.learningmodels.__name__)
                    logger.error("Did you load the necessary plugins (loaded: %s) and use the correct learningframework.kind (in use: '%s')?",
                                 config['bench'].get('plugins', []),
                                 learningframework.__class__.__name__)
                    raise ModuleNotFoundError


                trainresult = train(datamodel=data,
                        train_epochs=m.get('epochs', 0),
                        iteration=i,
                        model_name=m['name'],
                        model=model,
                        model_params=m.get('params', {}),
                        batch_size=m.get('batch_size', None),
                        optimizer=m.get('optimizer', None),
                        framework=learningframework,
                        load=m.get('load', False),
                        train=m.get('train', True),
                        evaluate=m.get('evaluate', True),
                        dataaugmentations=dataaugmentations,
                        experimenttracking=et)

                if et is not None:
                    et.stop()

                log(LearningModelLoggerFields(name=trainresult.name,
                                              i=i,
                                              params=trainresult.params,
                                              mem_params=trainresult.mem_params,
                                              accuracy=trainresult.acc))

                for postprocessing in config.get('postprocessing', []):
                    ppp = {k: v for k,v in postprocessing.get('params', {}).items()} # Workaround tomlkit bug where some nested dict would lose their items
                    trainresult, m = getattr(qualia.postprocessing, postprocessing['kind'])(**ppp)(
                        trainresult=trainresult,
                        model_conf=m)
                    if trainresult.log:
                        log(LearningModelLoggerFields(name=trainresult.name,
                                                      i=i,
                                                      params=trainresult.params,
                                                      mem_params=trainresult.mem_params,
                                                      accuracy=trainresult.acc))
                    if postprocessing.get('export', False):
                        trainresult.framework.export(trainresult.model, f'{trainresult.name}_r{i}')

        return loggers



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any

from qualia_core import random
from qualia_core.typing import TYPE_CHECKING

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    import numpy as np  # noqa: TCH002
    import numpy.typing  # noqa: TCH002

    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawDataModel  # noqa: TCH001
    from qualia_core.evaluation.Stats import Stats  # noqa: TCH001
    from qualia_core.learningframework.LearningFramework import LearningFramework  # noqa: TCH001

logger = logging.getLogger(__name__)

class Evaluator(ABC):
    def apply_dataaugmentations(self,
                                framework: LearningFramework[Any],
                                dataaugmentations: list[DataAugmentation] | None,
                                test_x: numpy.typing.NDArray[np.float32],
                                test_y: numpy.typing.NDArray[np.int32]) -> tuple[numpy.typing.NDArray[np.float32],
                                                                                 numpy.typing.NDArray[np.int32]]:
        """Apply evaluation :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation` to dataset.

        Only the :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation`
        with :attr:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation.evaluate` set are applied.
        This should not be used to apply actual data augmentation to the data, but rather use the conversion or transform
        :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation` modules.

        :param framework: The :class:`qualia_core.learningframework.LearningFramework.LearningFramework` instance
            providing the :meth:`qualia_core.learningframework.LearningFramework.LearningFramework.apply_dataaugmentation` method
            compatible with the provided :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation`
        :param dataaugmentations: List of :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation` objects
        :param test_x: Input data to apply data augmentation to
        :param test_y: Input labels to apply data augmentation to
        :return: Tuple of data and labels after
            applying :class:`qualia_core.dataaugmentation.DataAugmentation.DataAugmentation` sequentially
        """
        if dataaugmentations is None:
            return test_x, test_y
        for da in dataaugmentations:
            if da.evaluate:
                test_x, test_y = framework.apply_dataaugmentation(da, test_x, test_y)
        return test_x, test_y

    def shuffle_dataset(self,
                        test_x: numpy.typing.NDArray[np.float32],
                        test_y: numpy.typing.NDArray[np.int32]) -> tuple[numpy.typing.NDArray[np.float32],
                                                                         numpy.typing.NDArray[np.int32]]:
        """Shuffle the input data, keeping the labels in the same order as the shuffled data.

        Shuffling uses the seeded shared random generator from :obj:`qualia_core.random.shared`.

        :param test_x: Input data
        :param test_y: Input labels
        :return: Tuple of shuffled data and labels
        """
        perms = random.shared.generator.permutation(len(test_y))
        test_x = test_x[perms]
        test_y = test_y[perms]
        return test_x, test_y

    def limit_dataset(self,
                      test_x: numpy.typing.NDArray[np.float32],
                      test_y: numpy.typing.NDArray[np.int32],
                      limit: int | None) -> tuple[numpy.typing.NDArray[np.float32],
                                                  numpy.typing.NDArray[np.int32]]:
        """Truncate dataset to ``limit`` samples.

        :param test_x: Input data
        :param test_y: Input labels
        :param limit: Number of samples to truncate to, data is returned as-is if `None` or 0
        :return: Tuple of data and labels limited to ``limit`` samples
        """
        if limit:
            test_x = test_x[:limit]
            test_y = test_y[:limit]
        return test_x, test_y

    def compute_accuracy(self,
                         preds: list[int],
                         truth: numpy.typing.NDArray[np.int32]) -> float:
        """Compute accuracy from the target results.

        :param results: List of :class:`Result` from inference on the target
        :param test_y: Array of ground truth one-hot encoded
        :return: Accuracy (micro) between 0 and 1
        """
        correct = 0
        for line, pred in zip(truth, preds):
            logger.debug('%s %s', line.argmax(), pred)
            if line.argmax() == pred:
                correct += 1

        return correct / len(preds)

    @abstractmethod
    def evaluate(self,  # noqa: PLR0913
                 framework: LearningFramework[Any],
                 model_kind: str,
                 dataset: RawDataModel,
                 target: str,
                 tag: str,
                 limit: int | None = None,
                 dataaugmentations: list[DataAugmentation] | None = None) -> Stats | None:
        ...



from tensorflow.keras.layers import Conv1D as Conv
from tensorflow.keras.layers import AveragePooling1D as AveragePooling
from tensorflow.keras.layers import MaxPooling1D as MaxPooling
from tensorflow.keras.layers import ZeroPadding1D as ZeroPadding



import numpy as np

class FeatureSample:
    def __init__(self, features: 'List'=None):
        self.features = features if features is not None else []

    def get_raw_array(self):
        return np.reshape(self.features, (-1, 1))



import math
import numpy as np


import torch.nn as nn
import torch.nn.functional as F
from qualia_core.learningmodel.pytorch.layers import Add


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self,
                    layers_t,
                    in_planes,
                    planes,
                    kernel_size,
                    stride,
                    padding,
                    batch_norm,
                    bn_momentum,
                    force_projection_with_stride):
        


        super().__init__()
        self.batch_norm = batch_norm
        self.force_projection_with_stride = force_projection_with_stride
        self.in_planes = in_planes
        self.planes = planes
        self.stride = stride

        self.conv1 = layers_t.Conv(in_planes,
                                   planes,
                                   kernel_size=kernel_size,
                                   stride=1,
                                   padding=padding,
                                   bias=not batch_norm)
        if batch_norm:
            self.bn1 = layers_t.BatchNorm(planes, momentum=bn_momentum)
        if self.stride != 1:
            self.pool1 = layers_t.MaxPool(stride)
        self.relu1 = nn.ReLU()
        
        
        self.conv2 = layers_t.Conv(
            planes, planes, kernel_size=kernel_size, stride=1, padding=padding, bias=not batch_norm)
        if batch_norm:
            self.bn2 = layers_t.BatchNorm(planes, momentum=bn_momentum)
        if self.stride != 1:
            self.smax = layers_t.MaxPool(stride)
        if self.in_planes != self.expansion*self.planes or force_projection_with_stride and self.stride != 1:
            self.sconv = layers_t.Conv(in_planes, self.expansion*planes, kernel_size=1, stride=1, bias=not batch_norm)
            if batch_norm:
                self.sbn = layers_t.BatchNorm(
                    self.expansion*planes, momentum=bn_momentum)
        self.add = Add()
        self.relu = nn.ReLU()

    def forward(self, x):

        out = self.conv1(x)

        if self.batch_norm:
            out = self.bn1(out)
        if self.stride != 1:
            out = self.pool1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        if self.batch_norm:
            out = self.bn2(out)


        # shortcut
        tmp = x

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.sconv(tmp)

            if self.batch_norm:
                tmp = self.sbn(tmp)

        if self.stride != 1:
            tmp = self.smax(tmp)

        out = self.add(out, tmp)
        out = self.relu(out)
        return out

class ResNet(nn.Module):
    def __init__(self,
        input_shape,
        output_shape,
        filters: tuple=(15, 15, 30, 60, 120),
        kernel_sizes: tuple=(3, 3, 3, 3, 3),

        num_blocks: tuple=(2, 2, 2, 2),
        strides: tuple=(1, 1, 2, 2, 2),
        paddings: int=(1, 1, 1, 1, 1),
        prepool: int=1,
        postpool: str='max',
        batch_norm: bool=False,
        bn_momentum: float=0.1,
        force_projection_with_stride: bool=True,

        
        dims: int=1,
        basicblockbuilder=None):
        super().__init__()

        self.input_shape = input_shape
        self.output_shape = output_shape

        if dims == 1:
            import qualia_core.learningmodel.pytorch.layers1d as layers_t
        elif dims == 2:
            import qualia_core.learningmodel.pytorch.layers2d as layers_t
        else:
            raise ValueError('Only dims=1 or dims=2 supported')

        if basicblockbuilder is None:
            basicblockbuilder=lambda in_planes, planes, kernel_size, stride, padding: BasicBlock(
                layers_t=layers_t,
                in_planes=in_planes,
                planes=planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                batch_norm=batch_norm,
                bn_momentum=bn_momentum,
                force_projection_with_stride=force_projection_with_stride)

        self.in_planes = filters[0]
        self.batch_norm = batch_norm
        self.num_blocks = num_blocks
        if prepool > 1:
            self.prepool = layers_t.AvgPool(prepool)

        self.conv1 = layers_t.Conv(input_shape[-1], filters[0], kernel_size=kernel_sizes[0], stride=strides[0], padding=paddings[0], bias=not batch_norm)
        if self.batch_norm:
            self.bn1 = layers_t.BatchNorm(self.in_planes, momentum=bn_momentum)
        self.relu1 = nn.ReLU()

        self.layers = []
        for planes, kernel_size, stride, padding, num_block in zip(filters[1:], kernel_sizes[1:], strides[1:], paddings[1:], num_blocks):
            self.layers += [self._make_layer(basicblockbuilder, num_block, planes, kernel_size, stride, padding)]
        self.layers = nn.ModuleList(self.layers)

        # GlobalMaxPool kernel_size computation
        self._fm_dims = np.array(input_shape[:-1]) // np.array(prepool)
        for _, kernel, stride, padding in zip(filters, kernel_sizes, strides, paddings):
            self._fm_dims += np.array(padding) * 2
            self._fm_dims -= (kernel - 1)
            self._fm_dims = np.floor(self._fm_dims / stride).astype(int)

        if postpool == 'avg':
            self.postpool = layers_t.AdaptiveAvgPool(1)
        elif postpool == 'max':
            self.postpool = layers_t.MaxPool(tuple(self._fm_dims))

        self.flatten = nn.Flatten()
        self.linear = nn.Linear(self.in_planes*BasicBlock.expansion, output_shape[0])

    def _make_layer(self, basicblockbuilder, num_blocks, planes, kernel_size, stride, padding):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            block = basicblockbuilder(in_planes=self.in_planes, planes=planes, kernel_size=kernel_size, stride=stride, padding=padding)
            layers.append(block)
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        if hasattr(self, 'prepool'):
            x = self.prepool(x)

        out = self.conv1(x)
        if self.batch_norm:
            out = self.bn1(out)
        out = self.relu1(out)

        for i in range(len(self.layers)):
            out = self.layers[i](out)

        if hasattr(self, 'postpool'):
            out = self.postpool(out)

        out = self.flatten(out)
        out = self.linear(out)
        return out



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic

from qualia_core.learningframework.LearningFramework import LearningFramework, T
from qualia_core.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003, I001 # torch must be imported before keras to avoid deadlock

    from torch import nn  # noqa: TCH002
    import keras  # type: ignore[import] # No stubs for keras package  # noqa: TCH002
    import numpy.typing  # noqa: TCH002

class Converter(ABC, Generic[T]):
    deployers: ModuleType | None

    @abstractmethod
    def convert(self,
                framework: LearningFramework[nn.Module | keras.Model],
                model: nn.Module | keras.Model,
                model_name: str,
                representative_dataset: numpy.typing.NDArray[Any]) -> Converter[T] | None:
        ...

    def process_mem_params(self, mem_params: int) -> Callable[[LearningFramework[T],
                                                               T], int]:
        def f(_: LearningFramework[T], __: T) -> int:
            return mem_params
        return f



from __future__ import annotations

from pathlib import Path

from .STM32CubeAI import STM32CubeAI


class NucleoL452REP(STM32CubeAI):

    def __init__(self, outdir: Path | None = None) -> None:
        outdir = outdir if outdir is not None else Path('out')/'deploy'/'NucleoL452REP'

        super().__init__(outdir=outdir,
                        projectname='STM32CubeAI-NucleoL452REP',
                        projectdir=Path(__file__).parent.parent.parent/'assets'/'projects'/'stm32cubeai'/'NucleoL452REP')



from __future__ import annotations

import logging

from pydantic import TypeAdapter, ValidationError

from qualia_core.typing import TYPE_CHECKING, ConfigDict
from qualia_core.utils.merge_dict import merge_dict

if TYPE_CHECKING:
    from pathlib import Path  # noqa: TCH003

    from qualia_core.typing import RecursiveConfigDict

logger = logging.getLogger(__name__)

def validate_config_dict(config: RecursiveConfigDict) -> ConfigDict | None:
    ta = TypeAdapter(ConfigDict)
    try:
        validated_config: ConfigDict = ta.validate_python(config, strict=False)

        extra_dict_a_keys = [k for k in config if k not in validated_config]
        extra_dict_b_keys = [k for k in validated_config if k not in config]
        non_matching_values = {k: v for k, v in validated_config.items() if k in config and v != config[k]}

        if extra_dict_a_keys:
            logger.error('Missing keys after validation: %s', extra_dict_a_keys)

        if extra_dict_b_keys:
            logger.error('Extra keys after validation: %s', extra_dict_b_keys)

        if non_matching_values:
            for k, v in non_matching_values.items():
                logger.error('Different value after validation for key: %s', k)
                logger.error('Before validation:\n%s', config[k])
                logger.error('After validation:\n%s', v)

        if extra_dict_a_keys or extra_dict_b_keys or non_matching_values:
            return None

    except ValidationError:
        logger.exception('Error when validating configuration.')
        return None

    return validated_config

def parse_config(path: Path) -> tuple[RecursiveConfigDict, str]:
    import tomlkit

    with path.open() as f:
        toml_config = tomlkit.parse(f.read())

    # Convert to built-in Python types
    config: RecursiveConfigDict = toml_config.unwrap()

    # Merge settings from template into individual models
    if 'model_template' in config:
        models = config['model']
        model_template = config['model_template']

        if not isinstance(models, list):
            logger.error('`model` must be a list, got: %s', type(models))
            raise TypeError

        if not isinstance(model_template, dict):
            logger.error('`model_template` must be a dict, got: %s', type(model_template))
            raise TypeError

        for i, model in enumerate(models):
            if not isinstance(model, dict):
                logger.error('`model[%d]` must be a dict, got: %s', i, type(model))
                raise TypeError
            models[i] = merge_dict(model, model_template)

    return config, path.stem




"""Path utilities."""

from __future__ import annotations

import logging
import sys
from pathlib import Path

from qualia_core.typing import TYPE_CHECKING

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from importlib.abc import Traversable  # noqa: TCH003
    from importlib.readers import MultiplexedPath  # noqa: TCH003


logger = logging.getLogger(__name__)

def resources_to_path(resources: Path | MultiplexedPath | Traversable) -> Path:
    """Convert a (static type) Traversable to a Path if the underlying dynamic type is MultiplexedPath.

    This is useful when using :func:`importlib.resources.files` since with Python >= 3.10 it returns a
    :class:`importlib.readers.MultiplexedPath` when :class:`importlib.readers.NamespaceReader` is used as a reader,
    e.g. when used on a package available as a plain directory (not zip file). This allows getting a standard :class:`pathlib.Path`
    out of it.
    Any other situation will probably fail with ValueError.

    :param resources: An object returned by :func:`importlib.resources.files`
    :return: A standard Path object pointing to the ``resources`` path on the filesystem
    :raise ValueError: When conversion fails because ``resources`` was not a :class:`importlib.readers.MultiplexedPath` on Python
        >= 3.10 or a :class:`pathlib.Path`
    """
    if isinstance(resources, Path): # Already Path objected, no need for hackery
        return resources

    if sys.version_info >= (3, 10): # Python 3.10 may return MultiplexedPath
        from importlib.readers import MultiplexedPath
        if isinstance(resources, MultiplexedPath):
            return resources / '' # / operator applies to underlying Path

    logger.error('Could not convert %s to Path object: expected type Path or MultiplexedPath for Python >= 3.10, got %s',
                 resources,
                 type(resources))
    raise ValueError



from . import quantized_layers

from .Add import Add
from .CustomBatchNorm1d import CustomBatchNorm1d
from .CustomBatchNorm2d import CustomBatchNorm2d
from .GlobalSumPool1d import GlobalSumPool1d
from .GlobalSumPool2d import GlobalSumPool2d
from .QuantizedAdd import QuantizedAdd
from .QuantizedGlobalSumPool1d import QuantizedGlobalSumPool1d
from .QuantizedGlobalSumPool2d import QuantizedGlobalSumPool2d

__all__ = [
        'Add',
        'CustomBatchNorm1d',
        'CustomBatchNorm2d',
        'GlobalSumPool1d',
        'GlobalSumPool2d',
        'QuantizedAdd',
        'QuantizedGlobalSumPool1d',
        'QuantizedGlobalSumPool2d',
        ]

layers = (
        Add,
        CustomBatchNorm1d,
        CustomBatchNorm2d,
        QuantizedAdd,
        GlobalSumPool1d,
        GlobalSumPool2d,
        QuantizedGlobalSumPool1d,
        QuantizedGlobalSumPool2d,
        *quantized_layers.quantized_layers,
        )



import torch
from math import sin, pi
import torch.optim.lr_scheduler as sch

class SinDescent(sch._LRScheduler):
    def __init__(self, optimizer, epoch=-1, pme=0, w=1, lr0=0.1, lrf = 0.1, last_epoch=-1, verbose=False):
        self.epoch  = epoch
        self.pme    = pme
        self.w      = w
        self.lr0    = lr0
        self.lrf   = lrf
        super().__init__(optimizer, last_epoch, verbose)
    

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)
        return [self.sinD(self._step_count-1) for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        return [self.sinD(base_lr) for base_lr in self.base_lrs]
    
    def sinD(self, lr):
        return self.lts(lr)*self.sf(lr) + self.LGf(lr)
    
    def lts(self, lr):
        return (-lr/self.epoch + 1)* self.pme
    
    def sf(self, lr):
        return sin(lr/self.epoch *2*self.w*pi)
    
    def LGf(self, lr):
        ovs = (-self.epoch*self.lr0)/(self.lrf-self.lr0)
        return self.lr0*(-lr/ovs +1)



from .Preprocessing import Preprocessing

class RemoveSensor(Preprocessing):
    def __init__(self, sensorkinds, noconvert: bool=False):
        import qualia_core.datamodel.sensor as sensors
        for i, sensorkind in enumerate(sensorkinds):
            if not noconvert and isinstance(sensorkind, str):
                sensorkinds[i] = getattr(sensors, sensorkind)
        self.__excludesensors = sensorkinds

    def __call__(self, datamodel):
        for _, subjectset in datamodel:
            for subject in subjectset:
                for activity in subject.activities:
                    for sample in activity.samples:
                        sample.sensors = [sensor for sensor in sample.sensors if sensor.__class__ not in self.__excludesensors]
                    activity.samples = [sample for sample in activity.samples if len(sample.sensors) > 0] # Remove empty samples
        return datamodel



from __future__ import annotations

import sys
from pathlib import Path

import keras  # type: ignore[import-untyped]

from qualia_core.typing import TYPE_CHECKING

from .LearningFramework import LearningFramework

if TYPE_CHECKING:
    import tensorflow as tf  # type: ignore[import-untyped] # noqa: TCH002

    from qualia_core.dataaugmentation.DataAugmentation import DataAugmentation  # noqa: TCH001
    from qualia_core.datamodel.RawDataModel import RawData  # noqa: TCH001
    from qualia_core.experimenttracking.ExperimentTracking import ExperimentTracking  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class Keras(LearningFramework[keras.Model]):
    # Reference framework-specific external modules
    import qualia_core.learningmodel.keras as learningmodels
    import qualia_core.experimenttracking.pytorch as experimenttrackings

    def __init__(self):
        import tensorflow as tf
        # Cleanup, especially reset graph uid for uniquely generated layer names
        tf.keras.backend.clear_session()

    def __configure_optimizers(self, optimizer=None):
        import tensorflow as tf

        if optimizer is None:
            from .DummyOptimizer import DummyOptimizer
            return DummyOptimizer()

        opt = getattr(tf.keras.optimizers, optimizer['kind'])
        if 'scheduler' in optimizer:
            scheduler = getattr(tf.keras.optimizers.schedules, optimizer['scheduler']['kind'])(**optimizer['scheduler'].get('params', {}))
            opt = opt(learning_rate=scheduler, **optimizer.get('params', {}))
            print('Scheduler:', scheduler, optimizer['scheduler'].get('params', {}))
        else:
            opt = opt(**optimizer.get('params', {}))
        print('Optimizer:', opt, optimizer.get('params', {}))
        return opt


    def train(self,
              model,
              trainset,
              validationset,
              epochs,
              batch_size,
              optimizer,
              dataaugmentations=None,
              experimenttracking=None,
              name: str='') -> keras.Model:
        # gpus handled by TensorFlowInitializer
        import tensorflow as tf

        if dataaugmentations:
            raise ValueError(f'Data augmentation not supported with {self.__class__.__name__}')

        # You can plot the quantize training graph on tensorboard
        logdir='out/logs/fit/{name}/{datetime.now().strftime("%Y%m%d-%H%M%S")}'
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=logdir)

        callbacks=[]
        #callbacks.append(tensorboard_callback)
        # The patience parameter is the amount of epochs to check for improvement
        #callbacks.append(tf.keras.callbacks.EarlyStopping(monitor='val_accuracy', patience=5))
        #callbacks.append(tfdocs.modeling.EpochDots())
        if experimenttracking is not None:
            callbacks.append(experimenttracking.callback)

        optimizer = self.__configure_optimizers(optimizer)

        accuracy_metric = tf.keras.metrics.CategoricalAccuracy(name='trainacc')
        model.compile(loss='categorical_crossentropy', optimizer=optimizer, metrics=[accuracy_metric])

        model.fit(trainset.x,
                trainset.y,
                epochs=epochs,
                batch_size=batch_size,
                validation_data=validationset.astuple() if validationset is not None else None,
                callbacks=callbacks)

        # Remove softmax
        #compile_params = {'loss': model.loss, 'optimizer': model.optimizer, 'metrics': model.compiled_metrics.metrics}
        #model = tf.keras.Model(model.input, model.layers[-2].output, name=model.name)
        #model.compile(**compile_params)
        #model.summary()

        return model

    def load(self, name, model=None, path: Path=Path('out')/'learningmodel'):
        from tensorflow.keras.models import load_model
        model = load_model(path/f'{name}.h5', custom_objects=getattr(model, 'get_custom_objects', lambda: {})())
        return model

    def evaluate(self, model, testset, batch_size, gpus: int=None, dataaugmentations=None, experimenttracking=None, dataset_type: str='', name: str=''):
        import tensorflow as tf

        if dataaugmentations:
            raise ValueError(f'Data augmentation not supported with {self.__class__.__name__}')

        accuracy_metric = tf.keras.metrics.CategoricalAccuracy(name='testacc')
        model.compile(loss='categorical_crossentropy', metrics=[accuracy_metric])

        metrics = model.evaluate(testset.x, testset.y, batch_size=batch_size)

        # Predict and show confusion matrix
        train_predictions = model.predict(testset.x)

        cm = tf.math.confusion_matrix(testset.y.argmax(axis=1), train_predictions.argmax(axis=1))
        print(cm)
        if experimenttracking is not None:
            experimenttracking.logger.experiment['cm'].log(np.array2string(cm))

        metrics = {name: metric for name, metric in zip(model.metrics_names, metrics)}
        metrics['cm'] = cm

        return metrics

    @override
    def predict(self,  # noqa: PLR0913
                 model: keras.Model,
                 dataset: RawData,
                 batch_size: int,
                 dataaugmentations: list[DataAugmentation],
                 experimenttracking: ExperimentTracking | None,
                 name: str) -> tf.Tensor:
        raise NotImplementedError

    def export(self, model, name, path: Path=Path('out')/'learningmodel'):
        model.save(path/f'{name}.h5')

    def summary(self, model):
        model.summary()

    def n_params(self, model):
        return model.count_params()

    def save_graph_plot(self, model, model_save):
        import tensorflow as tf
        tf.keras.utils.model_to_dot(model, show_shapes=True, expand_nested=True).write(f'{model_save}.dot')

    def apply_dataaugmentation(self):
        raise NotImplemented()



import torch
import torch.nn as nn
from torch.fx import Tracer, GraphModule

from qualia_core.learningmodel.pytorch.layers import layers as custom_layers


class TorchVisionModel(nn.Module):
    # Custom tracer that generates call_module for our custom Qualia layers instead of attempting to trace their forward()
    class TracerCustomLayers(Tracer):
        def __init__(self, custom_layers: tuple, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.custom_layers = custom_layers

        def is_leaf_module(self, m: torch.nn.Module, module_qualified_name : str) -> bool:
            return super().is_leaf_module(m, module_qualified_name) or isinstance(m, custom_layers)

    def _shape_channels_last_to_first(self, shape):
        return (shape[-1], ) + shape[0:-1]

    def __init__(self, input_shape, output_shape, model, *args, **kwargs):
        import torchvision.models as models

        super().__init__()
        self.input_shape = input_shape
        self.output_shape = output_shape
        self.pretrained_model = getattr(models, model)(*args, **kwargs)

        self.fm = self.create_feature_extractor(self.pretrained_model, 'avgpool')
        for param in self.fm.parameters():
            param.requires_grad = False

        self.fm_shape = self.fm(torch.rand((1, *self._shape_channels_last_to_first(input_shape)))).shape

        print(tuple(self.fm_shape[1:]),self.output_shape)

        self.linear = nn.Linear(self.fm_shape[1], self.output_shape[0])
        self.flatten = nn.Flatten()

        for name, param in self.fm.named_parameters():
            print(name, param.requires_grad)

    # Similar to torchvision's but simplified for our specific use case
    def create_feature_extractor(self, model: nn.Module, return_node: str):
        # Feature extractor only used in eval mode
        model.eval()

        tracer = self.TracerCustomLayers(custom_layers=custom_layers)
        graph = tracer.trace(model)
        graph.print_tabular()
        graphmodule = GraphModule(tracer.root, graph, tracer.root.__class__.__name__)

        # Remove existing output node
        old_output = [n for n in graphmodule.graph.nodes if n.op == 'output']
        if not old_output:
            raise ValueError(f'No output in dl model')
        if len(old_output) > 1:
            raise ValueError(f'Multiple outputs in dl model')
        graphmodule.graph.erase_node(old_output[0])
        print(f'{old_output=}')

        # Find desired output layer
        new_output = [n for n in graphmodule.graph.nodes if n.name == return_node]
        if not new_output:
            raise ValueError(f'fm_output = \'{return_node}\' not found in dl model')
        if len(new_output) > 1:
            raise ValueError(f'Multiple matches for fm_output = \'{return_node}\' in dl model')

        # Add new output for desired layer
        with graphmodule.graph.inserting_after(list(graphmodule.graph.nodes)[-1]):
            graphmodule.graph.output(new_output[0])

        # Remove unused layers
        graphmodule.graph.eliminate_dead_code()

        graphmodule.recompile()

        graphmodule.graph.print_tabular()

        return graphmodule

    def forward(self, x):
        x = self.fm(x)
        x = self.flatten(x)
        x = self.linear(x)
        return x



from pathlib import Path

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from .Preprocessing import Preprocessing

class VisualizeWindows(Preprocessing):
    def __init__(self, labels: list):
        self.__labels = labels

    def __call__(self, datamodel):
        outpath = Path('out')/'visualization'/'windows'/datamodel.name

        fig = plt.figure()
        for sname, s in datamodel:
            outdir = outpath/sname
            outdir.mkdir(parents=True, exist_ok=True)

            pdffiles = {}
            for label in self.__labels:
                outfile = outdir/f'{label}.pdf'
                pdffiles[label] = PdfPages(outfile)

            for window_x, window_y, info in zip(s.x, s.y, s.info):
                label = np.bincount(window_y).argmax()

                for i in range(window_x.shape[-1]):
                    p = plt.plot(window_x[:,i])
                    for x, y, a in zip(range(window_x.shape[0]), window_x[:,i], window_y):
                        plt.annotate(a, (x, y))

                plt.suptitle('\n'.join([' '.join(inf) for inf in np.unique(info, axis=0)]))

                pdffiles[label].savefig(fig)
                fig.clf()

            for pdffile in pdffiles.values():
                pdffile.close()
        return datamodel



import importlib.util
import logging

from .Distribution import Distribution
from .Keras2TFLite import Keras2TFLite
from .PostProcessing import PostProcessing
from .QualiaCodeGen import QualiaCodeGen

__all__ = [
        'Distribution',
        'FuseBatchNorm',
        'Keras2TFLite',
        'PostProcessing',
        'QualiaCodeGen',
        ]

logger = logging.getLogger(__name__)

if importlib.util.find_spec('torch') is None:
    logger.warning('PyTorch is required for FuseBatchNorm, QuantizationAwareTraining, Torch2Keras')
else:
    from .FuseBatchNorm import FuseBatchNorm
    from .QuantizationAwareTraining import QuantizationAwareTraining

    __all__ += ['FuseBatchNorm',
                'QuantizationAwareTraining']

if importlib.util.find_spec('keras') is None:
    logger.warning('Keras is required for RemoveKerasSoftmax, Torch2Keras')
else:
    from .RemoveKerasSoftmax import RemoveKerasSoftmax

    __all__ += ['RemoveKerasSoftmax']

    # Warning message already printed
    if importlib.util.find_spec('torch') is not None:
        from .Torch2Keras import Torch2Keras

        __all__ += ['Torch2Keras']
